# Copyright 2023-2026 AgentEra(Agently.Tech)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from agently.utils import LazyImport

LazyImport.import_package("chromadb")

from typing import Literal, Callable, TypedDict, Any, TYPE_CHECKING, overload
from itertools import zip_longest

from chromadb import Client as ChromaDBClient
from chromadb.config import Settings
from chromadb.api.types import EmbeddingFunction

if TYPE_CHECKING:
    from chromadb.api.types import (
        Documents,
        Embedding,
        Embeddings,
        QueryResult,
        Schema,
        DataLoader,
        Loadable,
        Where,
        WhereDocument,
    )
    from chromadb.api.collection_configuration import CreateCollectionConfiguration
    from chromadb.api import ClientAPI
    from agently.base import Agent


class ChromaDataDictOptional(TypedDict, total=False):
    metadata: dict[Any, Any]
    id: Any
    embedding: "Embedding"


class ChromaDataDict(ChromaDataDictOptional):
    document: str


class ChromaData:
    def __init__(
        self,
        original_data: ChromaDataDict | list[ChromaDataDict],
        *,
        embedding_function: "Callable[[str | list[str]], Embeddings] | None" = None,
        agent: "Agent | None" = None,
    ):
        self._original_data = original_data if isinstance(original_data, list) else [original_data]
        if embedding_function:
            self._embedding_function = embedding_function
        elif agent:

            def embedding_function_by_agent(texts: str | list[str]) -> "Embeddings":
                return agent.input(texts).start()

            self._embedding_function = embedding_function_by_agent
        else:
            self._embedding_function = None
        self._chroma_data = {
            "documents": [],
            "metadatas": [],
            "ids": [],
            "embeddings": [],
        }
        if self._embedding_function is not None:
            documents = [data["document"] for data in self._original_data]
            embeddings = self._embedding_function(documents)
            self._chroma_data["embeddings"] = embeddings
            for index, data in enumerate(self._original_data):
                data["embedding"] = embeddings[index]
        else:
            del self._chroma_data["embeddings"]
        for index, data in enumerate(self._original_data):
            self._chroma_data["documents"].append(data["document"])
            self._chroma_data["metadatas"].append(data["metadata"] if "metadata" in data else None)
            self._chroma_data["ids"].append(data["id"] if "id" in data else str(index))

    def get_kwargs(self):
        return self._chroma_data

    def get_original_data(self):
        return self._original_data


class ChromaResults:
    def __init__(
        self,
        *,
        queries: str | list[str],
        results: "QueryResult",
        distance: float | None = None,
    ):
        if isinstance(queries, str):
            queries = [queries]

        ids = results.get("ids") or []
        documents = results.get("documents")
        metadatas = results.get("metadatas")
        distances = results.get("distances")

        if ids and len(queries) != len(ids):
            raise ValueError(
                f"The length of queries does not equal the length of results['ids'].\nQueries: {queries}\nIds: {ids}"
            )

        self._chroma_results = results
        self._results: dict[str, list[dict]] = {}

        doc_iter = documents if documents is not None else []
        meta_iter = metadatas if metadatas is not None else []
        dist_iter = distances if distances is not None else []

        for query, row_ids, row_documents, row_metadatas, row_distances in zip_longest(
            queries, ids, doc_iter, meta_iter, dist_iter, fillvalue=None
        ):
            row_ids = row_ids or []
            query_results: list[dict[str, Any]] = []
            for id_, doc, meta, dist in zip_longest(
                row_ids, row_documents or [], row_metadatas or [], row_distances or [], fillvalue=None
            ):
                query_result = {"id": id_, "document": doc, "metadata": meta, "distance": dist}
                if distance is None:
                    query_results.append(query_result)
                else:
                    if isinstance(dist, (int, float)) and dist < distance:
                        query_results.append(query_result)

            if query:
                self._results[query] = query_results

    def get_original_results(self):
        return self._chroma_results

    def get(self, *, simplify_single_result: bool = False):
        if simplify_single_result:
            results = list(self._results.values())
            if len(results) == 1:
                return results[0]
        return self._results


class ChromaEmbeddingFunction(EmbeddingFunction):
    def __init__(
        self,
        *,
        embedding_agent: "Agent",
    ):
        def embedding_function_by_agent(texts: list[str]) -> "Embeddings":
            return embedding_agent.input(texts).start()

        self.embedding_function = embedding_function_by_agent

    def __call__(self, documents: "Documents | list[str] | str") -> "Embeddings":
        if isinstance(documents, str):
            documents = [documents]
        return self.embedding_function([document for document in documents])


class ChromaCollection:

    def __init__(
        self,
        collection_name: str,
        *,
        conn: "ClientAPI | None" = None,
        schema: "Schema | None" = None,
        configuration: "CreateCollectionConfiguration | None" = None,
        metadata: dict[str, Any] | None = None,
        embedding_agent: "Agent | None" = None,
        data_loader: "DataLoader[Loadable] | None" = None,
        get_or_create: bool = False,
        hnsw_space: Literal["l2", "cosine", "ip"] = "cosine",
    ):
        if conn is None:
            conn = ChromaDBClient(Settings(anonymized_telemetry=False))

        if metadata is None:
            metadata = {}

        metadata["hnsw:space"] = hnsw_space

        self.embedding_function = (
            ChromaEmbeddingFunction(embedding_agent=embedding_agent) if embedding_agent is not None else None
        )

        self._collection = conn.create_collection(
            collection_name,
            schema=schema,
            configuration=configuration,
            metadata=metadata,
            embedding_function=self.embedding_function,
            data_loader=data_loader,
            get_or_create=get_or_create,
        )

    def add(
        self,
        data: ChromaDataDict | list[ChromaDataDict] | ChromaData,
        *,
        embedding_function: "Callable[[str | list[str]], Embeddings] | None" = None,
    ):
        embedding_function = embedding_function or self.embedding_function
        if embedding_function is None:
            raise NotImplementedError(
                f"Embedding function must be assigned either in ChromaCollection initialize or .add() method."
            )
        if not isinstance(data, ChromaData):
            data = ChromaData(data, embedding_function=embedding_function)
        self._collection.add(**data.get_kwargs())
        return self

    @overload
    def query(
        self,
        query_or_queries: str,
        *,
        embedding_function: "Callable[[str | list[str]], Embeddings] | None" = None,
        top_n: int | None = None,
        where: "Where | None" = None,
        where_document: "WhereDocument | None" = None,
        distance: float | None = None,
    ) -> list[dict[str, Any]]: ...

    @overload
    def query(
        self,
        query_or_queries: list[str],
        *,
        embedding_function: "Callable[[str | list[str]], Embeddings] | None" = None,
        top_n: int | None = None,
        where: "Where | None" = None,
        where_document: "WhereDocument | None" = None,
        distance: float | None = None,
    ) -> dict[str, list[dict[str, Any]]]: ...

    def query(
        self,
        query_or_queries: str | list[str],
        *,
        embedding_function: "Callable[[str | list[str]], Embeddings] | None" = None,
        top_n: int | None = None,
        where: "Where | None" = None,
        where_document: "WhereDocument | None" = None,
        distance: float | None = None,
    ):
        embedding_function = embedding_function or self.embedding_function
        if embedding_function is None:
            raise NotImplementedError(
                f"Embedding function must be assigned either in ChromaCollection initialize or .query() method."
            )
        if not isinstance(query_or_queries, list):
            query_or_queries = [query_or_queries]
        return ChromaResults(
            queries=query_or_queries,
            results=self._collection.query(
                query_embeddings=embedding_function(query_or_queries),
                n_results=top_n or 10,
                where=where,
                where_document=where_document,
            ),
            distance=distance,
        ).get(simplify_single_result=True)

    def query_embeddings(
        self,
        embeddings_dict: dict[str, "Embedding"],
        *,
        top_n: int | None = None,
        where: "Where | None" = None,
        where_document: "WhereDocument | None" = None,
        distance: float | None = None,
    ):
        queries = [query for query in embeddings_dict.keys()]
        embeddings = [embeddings for embeddings in embeddings_dict.values()]
        return ChromaResults(
            queries=queries,
            results=self._collection.query(
                query_embeddings=embeddings,
                n_results=top_n or 10,
                where=where,
                where_document=where_document,
            ),
            distance=distance,
        ).get(simplify_single_result=True)



# Copyright 2023-2026 AgentEra(Agently.Tech)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import uuid
import yaml
from warnings import warn

from typing import Any, Literal, Sequence, TYPE_CHECKING, cast

from agently.core import BaseAgent
from agently.types.data import ChatMessage, ChatMessageDict
from agently.utils import RuntimeData, DataPathBuilder

if TYPE_CHECKING:
    from agently.types.data import ChatMessage, AgentlyModelResult, PromptStandardSlot
    from agently.core import Prompt
    from agently.core.ModelRequest import ModelResponseResult
    from agently.utils import Settings


class ChatSessionExtension(BaseAgent):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._activated_chat_session = None
        self.settings.setdefault("record_input_paths", [], inherit=True)
        self.settings.setdefault("record_input_mode", "all", inherit=True)
        self.settings.setdefault("record_output_paths", [], inherit=True)
        self.settings.setdefault("record_output_mode", "all", inherit=True)
        self.chat_session_runtime = RuntimeData()

        self.extension_handlers.append("finally", self.__finally)

        warn(
            "[Agently Extensions] ChatSessionExtension is deprecated and will be removed in future version, use SessionExtension which is mixed in by default instead."
        )

    def activate_chat_session(self, chat_session_id: str | None = None):
        if chat_session_id is None:
            chat_session_id = uuid.uuid4().hex
        self._activated_chat_session = chat_session_id
        chat_history = self.chat_session_runtime.get(chat_session_id, [])
        self.set_chat_history(chat_history)
        return chat_session_id

    def stop_chat_session(self):
        self._activated_chat_session = None
        self.set_chat_history([])
        return self

    def set_record_input_paths(
        self,
        *prompt_keys_and_paths: "PromptStandardSlot | tuple[PromptStandardSlot, str | None]",
        style: Literal["dot", "slash"] = "dot",
    ):
        record_input_paths = []
        for prompt_key_and_path in prompt_keys_and_paths:
            if isinstance(prompt_key_and_path, str):
                path = (prompt_key_and_path, None)
            else:
                if style == "slash":
                    path = (
                        prompt_key_and_path[0],
                        (
                            DataPathBuilder.convert_slash_to_dot(prompt_key_and_path[1])
                            if prompt_key_and_path[1] is not None
                            else None
                        ),
                    )
                else:
                    path = prompt_key_and_path
            if path not in record_input_paths:
                record_input_paths.append(path)
        self.settings.set("record_input_paths", record_input_paths)
        return self

    def add_record_input_paths(
        self,
        *prompt_keys_and_paths: "PromptStandardSlot | tuple[PromptStandardSlot, str | None]",
        style: Literal["dot", "slash"] = "dot",
    ):
        record_input_paths = self.settings.get("record_input_paths", [])
        assert isinstance(record_input_paths, list)
        for prompt_key_and_path in prompt_keys_and_paths:
            if isinstance(prompt_key_and_path, str):
                path = (prompt_key_and_path, None)
            else:
                if style == "slash":
                    path = (
                        prompt_key_and_path[0],
                        (
                            DataPathBuilder.convert_slash_to_dot(prompt_key_and_path[1])
                            if prompt_key_and_path[1] is not None
                            else None
                        ),
                    )
                else:
                    path = prompt_key_and_path
            if path not in record_input_paths:
                record_input_paths.append(path)
        self.settings.set("record_input_paths", record_input_paths)
        return self

    def add_record_input_path(
        self,
        prompt_key: "PromptStandardSlot",
        path: str | None = None,
        *,
        style: Literal["dot", "slash"] = "dot",
    ):
        record_input_paths = self.settings.get("record_input_paths", [])
        assert isinstance(record_input_paths, list)
        if path is not None and style == "slash":
            path = DataPathBuilder.convert_slash_to_dot(path)
        if (prompt_key, path) not in record_input_paths:
            record_input_paths.append((prompt_key, path))
        self.settings.set("record_input_paths", record_input_paths)
        return self

    def get_record_input_paths(self, *, style: Literal["dot", "slash"] = "dot"):
        if style == "dot":
            return self.settings.get("record_input_paths")
        else:
            result = []
            record_input_paths = self.settings.get("record_input_paths", [])
            if isinstance(record_input_paths, list):
                for path in record_input_paths:
                    if isinstance(path, str):
                        result.append(DataPathBuilder.convert_dot_to_slash(path))
            return result

    def clean_record_input_paths(self):
        self.settings.set("record_input_paths", [])
        return self

    def set_record_output_paths(
        self,
        *paths: str,
        style: Literal["dot", "slash"] = "dot",
    ):
        record_output_paths = []
        for path in paths:
            if style == "slash":
                path = DataPathBuilder.convert_slash_to_dot(path)
            if path not in record_output_paths:
                record_output_paths.append(path)
        self.settings.set("record_output_paths", record_output_paths)
        return self

    def add_record_output_paths(
        self,
        *paths: str,
        style: Literal["dot", "slash"] = "dot",
    ):
        record_output_paths = self.settings.get("record_output_paths", [])
        assert isinstance(record_output_paths, list)
        for path in paths:
            if style == "slash":
                path = DataPathBuilder.convert_slash_to_dot(path)
            if path not in record_output_paths:
                record_output_paths.append(path)
        self.settings.set("record_output_paths", record_output_paths)
        return self

    def add_record_output_path(
        self,
        path: str,
        *,
        style: Literal["dot", "slash"] = "dot",
    ):
        record_output_paths = self.settings.get("record_output_paths", [])
        assert isinstance(record_output_paths, list)
        if style == "slash":
            path = DataPathBuilder.convert_slash_to_dot(path)
        if path not in record_output_paths:
            record_output_paths.append(path)
        self.settings.set("record_output_paths", record_output_paths)
        return self

    def clean_record_output_paths(self):
        self.settings.set("record_output_paths", [])
        return self

    def get_record_output_paths(self, *, style: Literal["dot", "slash"] = "dot"):
        if style == "dot":
            return self.settings.get("record_output_paths")
        else:
            result = []
            record_output_paths = self.settings.get("record_output_paths", [])
            if isinstance(record_output_paths, list):
                for path in record_output_paths:
                    if isinstance(path, str):
                        result.append(DataPathBuilder.convert_dot_to_slash(path))
            return result

    def set_record_input_mode(self, mode: Literal["first", "all"]):
        self._record_input_mode = mode
        return self

    def set_record_output_mode(self, mode: Literal["first", "all"]):
        self._record_output_mode = mode
        return self

    def add_chat_history(self, chat_history: Sequence[ChatMessageDict | ChatMessage] | ChatMessageDict | ChatMessage):
        super().add_chat_history(chat_history)
        if self._activated_chat_session is not None:
            self.chat_session_runtime.set(self._activated_chat_session, self.prompt.get("chat_history"))
        return self

    async def __finally(self, result: "ModelResponseResult", settings: "Settings"):
        # Find user chat content to record
        record_input_paths = settings.get("record_input_paths")
        record_input_mode = settings.get("record_input_mode")
        if isinstance(record_input_paths, str):
            record_input_paths = [record_input_paths]
        if not isinstance(record_input_paths, list) or len(record_input_paths) == 0:
            self.add_chat_history({"role": "user", "content": result.prompt.to_messages()[-1]["content"]})
        else:
            record_input_paths = cast(
                "list[tuple[PromptStandardSlot, str | None]]",
                record_input_paths,
            )
            content = {}
            for prompt_key, path in record_input_paths:
                if path is None:
                    value = result.prompt.get(prompt_key)
                    if record_input_mode == "first":
                        content = str(value)
                        break
                    content[prompt_key] = value
                else:
                    prompt_value = result.prompt.get(prompt_key)
                    if isinstance(prompt_value, dict):
                        path_value = DataPathBuilder.get_value_by_path(prompt_value, path)
                        if path_value:
                            if record_input_mode == "first":
                                content = path_value
                                break
                            if prompt_key not in content or not isinstance(content[prompt_key], dict):
                                content[prompt_key] = {}
                            content[prompt_key][path] = path_value
            self.add_chat_history(
                {
                    "role": "user",
                    "content": yaml.safe_dump(content) if isinstance(content, dict) else content,
                }
            )
        # Find assistant chat content to record
        record_output_paths = self.settings.get("record_output_paths")
        record_output_mode = self.settings.get("record_output_mode")
        if not isinstance(record_output_paths, list) or len(record_output_paths) == 0:
            self.add_chat_history(
                {
                    "role": "assistant",
                    "content": result.full_result_data["text_result"],
                }
            )
            return
        if isinstance(record_output_paths, list):
            original_result = result.full_result_data["parsed_result"]
            content = {}
            if isinstance(original_result, dict):
                for path in record_output_paths:
                    path_value = DataPathBuilder.get_value_by_path(original_result, str(path))
                    if path_value:
                        if record_output_mode == "first":
                            self.add_chat_history(
                                {
                                    "role": "assistant",
                                    "content": path_value,
                                }
                            )
                            return
                        content[path] = path_value
                self.add_chat_history(
                    {
                        "role": "assistant",
                        "content": yaml.safe_dump(content),
                    }
                )
                return
            self.add_chat_history(
                {
                    "role": "assistant",
                    "content": result.full_result_data["text_result"],
                }
            )



# Copyright 2023-2026 AgentEra(Agently.Tech)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import uuid
import asyncio
import warnings
import json
import yaml
from pathlib import Path
from json import JSONDecodeError
from contextvars import ContextVar

from typing import Any, Literal, TYPE_CHECKING

if TYPE_CHECKING:
    from .TriggerFlow import TriggerFlow
    from agently.types.trigger_flow import TriggerFlowAllHandlers
    from agently.types.data import SerializableValue

from agently.utils import RuntimeData, FunctionShifter, GeneratorConsumer, Settings
from agently.types.trigger_flow import TriggerFlowRuntimeData, RUNTIME_STREAM_STOP
from agently.types.data import EMPTY
from .Control import (
    TriggerFlowPauseSignal,
    TRIGGER_FLOW_STATUS_CANCELLED,
    TRIGGER_FLOW_STATUS_COMPLETED,
    TRIGGER_FLOW_STATUS_CREATED,
    TRIGGER_FLOW_STATUS_FAILED,
    TRIGGER_FLOW_STATUS_RUNNING,
    TRIGGER_FLOW_STATUS_WAITING,
)
from .Signal import TriggerFlowSignal, TriggerFlowSignalType


class TriggerFlowExecution:
    def __init__(
        self,
        *,
        handlers: "TriggerFlowAllHandlers",
        trigger_flow: "TriggerFlow",
        id: str | None = None,
        skip_exceptions: bool = False,
        concurrency: int | None = None,
    ):
        # Basic Attributions
        self.id = id if id is not None else uuid.uuid4().hex
        self._handlers = handlers
        self._trigger_flow = trigger_flow
        self._runtime_data = RuntimeData()
        self._runtime_resources = RuntimeData(
            name=f"TriggerFlowExecution-{ self.id }-RuntimeResources",
            parent=self._trigger_flow._runtime_resources,
        )
        self._system_runtime_data = RuntimeData()
        self._skip_exceptions = skip_exceptions
        self._concurrency_semaphore = asyncio.Semaphore(concurrency) if concurrency and concurrency > 0 else None
        self._concurrency_depth = ContextVar(
            f"trigger_flow_execution_concurrency_depth_{ self.id }",
            default=0,
        )

        # Settings
        self.settings = Settings(
            parent=self._trigger_flow.settings,
            name=f"TriggerFlowExecution-{ self.id }-Settings",
        )

        # Emit
        self.emit = FunctionShifter.syncify(self.async_emit)

        # Flow Data
        self.get_flow_data = self._trigger_flow.get_flow_data
        self.set_flow_data = self._trigger_flow.set_flow_data
        self.async_set_flow_data = self._trigger_flow.async_set_flow_data
        self.append_flow_data = self._trigger_flow.append_flow_data
        self.async_append_flow_data = self._trigger_flow.async_append_flow_data
        self.del_flow_data = self._trigger_flow.del_flow_data
        self.async_del_flow_data = self._trigger_flow.async_del_flow_data

        # Runtime Data
        self.get_runtime_data = self._runtime_data.get
        self.set_runtime_data = FunctionShifter.syncify(self.async_set_runtime_data)
        self.append_runtime_data = FunctionShifter.syncify(self.async_append_runtime_data)
        self.del_runtime_data = FunctionShifter.syncify(self.async_del_runtime_data)
        self.set_runtime_resource = self._set_runtime_resource
        self.get_runtime_resource = self._get_runtime_resource
        self.del_runtime_resource = self._del_runtime_resource
        self.update_runtime_resources = self._update_runtime_resources
        self.clear_runtime_resources = self._clear_runtime_resources

        # Start
        self.start = FunctionShifter.syncify(self.async_start)

        # Runtime Stream
        self.put_into_stream = FunctionShifter.syncify(self.async_put_into_stream)
        self.stop_stream = FunctionShifter.syncify(self.async_stop_stream)

        # Pause / Continue
        self.pause_for = FunctionShifter.syncify(self.async_pause_for)
        self.continue_with = FunctionShifter.syncify(self.async_continue_with)

        # Result
        self.get_result = FunctionShifter.syncify(self.async_get_result)

        # Execution Status
        self._started = False
        self._status = TRIGGER_FLOW_STATUS_CREATED
        self._system_runtime_data.set("status", self._status)
        self._system_runtime_data.set("interrupts", {})
        self._system_runtime_data.set("last_signal", None)
        self._system_runtime_data.set("result", EMPTY)
        self._system_runtime_data.set("result_ready", asyncio.Event())
        self._runtime_stream_queue = asyncio.Queue()
        self._runtime_stream_consumer: GeneratorConsumer | None = None

    def _to_serializable_value(self, value: Any):
        return json.loads(RuntimeData({"value": value}).dump("json"))["value"]

    def _set_status(self, status: str):
        self._status = status
        self._system_runtime_data.set("status", status)

    def get_status(self):
        return self._status

    def is_waiting(self):
        return self._status == TRIGGER_FLOW_STATUS_WAITING

    def _get_interrupts(self) -> dict[str, Any]:
        interrupts = self._system_runtime_data.get("interrupts", {}, inherit=False)
        return interrupts if isinstance(interrupts, dict) else {}

    def get_interrupt(self, interrupt_id: str):
        return self._get_interrupts().get(interrupt_id)

    def get_pending_interrupts(self):
        return {
            interrupt_id: interrupt
            for interrupt_id, interrupt in self._get_interrupts().items()
            if isinstance(interrupt, dict) and interrupt.get("status") == "waiting"
        }

    def _set_runtime_resource(self, key: str, value: Any):
        self._runtime_resources.set(str(key), value)
        return self

    def _get_runtime_resource(self, key: str, default: Any = None):
        return self._runtime_resources.get(str(key), default)

    def require_runtime_resource(self, key: str):
        key = str(key)
        if key not in self._runtime_resources:
            available = sorted(str(resource_key) for resource_key in self.get_runtime_resources().keys())
            raise KeyError(
                f"Execution { self.id } missing required runtime resource '{ key }'. "
                f"Available resources: { available }"
            )
        return self._runtime_resources.get(key)

    def _del_runtime_resource(self, key: str):
        self._runtime_resources.pop(str(key), None)
        return self

    def _update_runtime_resources(
        self,
        mapping: dict[str, Any] | None = None,
        **kwargs,
    ):
        if mapping is not None:
            for key, value in dict(mapping).items():
                self._set_runtime_resource(str(key), value)
        for key, value in kwargs.items():
            self._set_runtime_resource(str(key), value)
        return self

    def _clear_runtime_resources(self):
        self._runtime_resources.clear()
        return self

    def get_runtime_resources(self):
        resources = self._runtime_resources.get(None, {}, inherit=True)
        return resources if isinstance(resources, dict) else {}

    def _serialize_signal(self, signal: TriggerFlowSignal | dict[str, Any] | None):
        if signal is None:
            return None
        if isinstance(signal, TriggerFlowSignal):
            return self._to_serializable_value(signal.to_state_dict())
        return self._to_serializable_value(signal)

    def _restore_signal(self, signal_state: dict[str, Any] | None):
        if not isinstance(signal_state, dict):
            return None
        try:
            return TriggerFlowSignal(
                id=str(signal_state.get("id")),
                trigger_event=str(signal_state.get("trigger_event")),
                trigger_type=signal_state.get("trigger_type", "event"),
                value=signal_state.get("value"),
                layer_marks=list(signal_state.get("layer_marks", [])),
                source=str(signal_state.get("source", "runtime")),
                meta=dict(signal_state.get("meta", {})),
            )
        except Exception:
            return None

    def _build_signal(
        self,
        trigger_event: str,
        value: Any = None,
        _layer_marks: list[str] | None = None,
        *,
        trigger_type: TriggerFlowSignalType = "event",
        source: str = "runtime",
        meta: dict[str, Any] | None = None,
    ):
        return TriggerFlowSignal.create(
            trigger_event=trigger_event,
            trigger_type=trigger_type,
            value=value,
            layer_marks=_layer_marks,
            source=source,
            meta=meta,
        )

    def _remember_signal(self, signal: TriggerFlowSignal):
        self._system_runtime_data.set("last_signal", signal.to_state_dict())

    def get_last_signal(self):
        return self._restore_signal(self._system_runtime_data.get("last_signal", None, inherit=False))

    def save(
        self,
        path: str | Path | None = None,
        *,
        encoding: str | None = "utf-8",
    ):
        result = self._system_runtime_data.get("result")
        result_ready = result is not EMPTY
        state = {
            "execution_id": self.id,
            "status": self._status,
            "runtime_data": json.loads(self._runtime_data.dump("json")),
            "flow_data": json.loads(self._trigger_flow._flow_data.dump("json")),
            "interrupts": self._to_serializable_value(self._get_interrupts()),
            "last_signal": self._serialize_signal(self.get_last_signal()),
            "resource_keys": sorted(str(key) for key in self.get_runtime_resources().keys()),
            "result": {
                "ready": result_ready,
                "value": self._to_serializable_value(result) if result_ready else None,
            },
        }
        if path is None:
            return state

        target = Path(path)
        suffix = target.suffix.lower()
        if suffix in {".yaml", ".yml"}:
            content = yaml.safe_dump(
                state,
                indent=2,
                allow_unicode=True,
                sort_keys=False,
            )
        else:
            content = json.dumps(
                state,
                indent=2,
                ensure_ascii=False,
            )
        target.parent.mkdir(parents=True, exist_ok=True)
        target.write_text(content, encoding=encoding)
        return state

    def load(
        self,
        state: dict[str, Any] | str | Path,
        *,
        encoding: str | None = "utf-8",
        runtime_resources: dict[str, Any] | None = None,
    ):
        if isinstance(state, (str, Path)):
            path = Path(state)
            is_file = False
            try:
                is_file = path.exists() and path.is_file()
            except (OSError, ValueError):
                is_file = False
            if is_file:
                suffix = path.suffix.lower()
                content = path.read_text(encoding=encoding)
                if suffix in {".yaml", ".yml"}:
                    try:
                        state = yaml.safe_load(content)
                    except yaml.YAMLError as e:
                        raise ValueError(f"Can not load TriggerFlowExecution state from YAML file '{ state }'.\nError: { e }")
                else:
                    try:
                        state = json.loads(content)
                    except JSONDecodeError as e:
                        raise ValueError(f"Can not load TriggerFlowExecution state from JSON file '{ state }'.\nError: { e }")
            elif isinstance(state, str):
                original = state
                try:
                    state = json.loads(state)
                except JSONDecodeError:
                    try:
                        state = yaml.safe_load(state)
                    except yaml.YAMLError as e:
                        raise ValueError(
                            f"Can not load TriggerFlowExecution state from JSON/YAML content.\nError: { e }\nContent: { original }"
                        )
            else:
                raise TypeError(f"Can not load TriggerFlowExecution state, expect dictionary/string/path but got: { type(state) }")

        if state is None:
            raise TypeError("Can not load TriggerFlowExecution state, got None.")

        if not isinstance(state, dict):
            raise TypeError(f"Can not load TriggerFlowExecution state, expect dictionary but got: { type(state) }")

        runtime_data = state.get("runtime_data", {})
        if not isinstance(runtime_data, dict):
            raise TypeError(f"Can not load key 'runtime_data', expect dictionary but got: { type(runtime_data) }")

        flow_data = state.get("flow_data", {})
        if not isinstance(flow_data, dict):
            raise TypeError(f"Can not load key 'flow_data', expect dictionary but got: { type(flow_data) }")

        interrupts = state.get("interrupts", {})
        if not isinstance(interrupts, dict):
            raise TypeError(f"Can not load key 'interrupts', expect dictionary but got: { type(interrupts) }")

        last_signal_state = state.get("last_signal", None)
        if last_signal_state is not None and not isinstance(last_signal_state, dict):
            raise TypeError(f"Can not load key 'last_signal', expect dictionary/None but got: { type(last_signal_state) }")

        result_state = state.get("result", {})
        if not isinstance(result_state, dict):
            raise TypeError(f"Can not load key 'result', expect dictionary but got: { type(result_state) }")

        ready = bool(result_state.get("ready", False))
        result_value = result_state.get("value")
        status = str(state.get("status", TRIGGER_FLOW_STATUS_CREATED))

        self._runtime_data.clear()
        self._runtime_data.update(runtime_data)

        self._trigger_flow._flow_data.clear()
        self._trigger_flow._flow_data.update(flow_data)

        result_ready = asyncio.Event()
        if ready:
            self._system_runtime_data.set("result", result_value)
            result_ready.set()
        else:
            self._system_runtime_data.set("result", EMPTY)
        self._system_runtime_data.set("result_ready", result_ready)
        self._system_runtime_data.set("interrupts", interrupts)
        self._system_runtime_data.set("last_signal", last_signal_state)
        self._set_status(status)
        self._started = status != TRIGGER_FLOW_STATUS_CREATED or bool(runtime_data) or ready or bool(interrupts)
        if runtime_resources:
            self.update_runtime_resources(runtime_resources)

        return self

    # Set Settings
    def set_settings(self, key: str, value: "SerializableValue"):
        self.settings.set_settings(key, value)
        return self

    # Set Concurrency
    def set_concurrency(self, concurrency):
        self._concurrency_semaphore = asyncio.Semaphore(concurrency) if concurrency and concurrency > 0 else None
        return self

    # Emit Event
    async def async_emit(
        self,
        trigger_event: str,
        value: Any = None,
        _layer_marks: list[str] | None = None,
        *,
        trigger_type: Literal["event", "runtime_data", "flow_data"] = "event",
        _source: str = "runtime",
        _meta: dict[str, Any] | None = None,
    ):
        signal = self._build_signal(
            trigger_event,
            value,
            _layer_marks,
            trigger_type=trigger_type,
            source=_source,
            meta=_meta,
        )
        return await self._async_dispatch_signal(signal)

    async def _async_dispatch_signal(self, signal: TriggerFlowSignal):
        from agently.base import async_system_message

        self._remember_signal(signal)
        await async_system_message(
            "TRIGGER_FLOW",
            signal.to_debug_dict(),
            self.settings,
        )
        tasks = []
        handlers = self._handlers[signal.trigger_type]

        if signal.trigger_event in handlers:
            for handler_id, handler in handlers[signal.trigger_event].items():
                await async_system_message(
                    "TRIGGER_FLOW",
                    {
                        "EVENT": signal.trigger_event,
                        "TYPE": signal.trigger_type,
                        "HANDLER": handler_id,
                        "SIGNAL_ID": signal.id,
                    },
                    self.settings,
                )

                async def run_handler(handler_func):
                    if self._concurrency_semaphore is None:
                        return await handler_func
                    depth = self._concurrency_depth.get()
                    token = self._concurrency_depth.set(depth + 1)
                    try:
                        if depth > 0:
                            return await handler_func
                        async with self._concurrency_semaphore:
                            return await handler_func
                    finally:
                        self._concurrency_depth.reset(token)

                handler_task = FunctionShifter.asyncify(handler)(
                    TriggerFlowRuntimeData(
                        trigger_event=signal.trigger_event,
                        trigger_type=signal.trigger_type,
                        value=signal.value,
                        execution=self,
                        _layer_marks=signal.layer_marks.copy(),
                        signal=signal,
                    )
                )
                tasks.append(asyncio.ensure_future(run_handler(handler_task)))

        if tasks:
            try:
                await asyncio.gather(*tasks, return_exceptions=self._skip_exceptions)
            except Exception:
                self._set_status(TRIGGER_FLOW_STATUS_FAILED)
                raise

    # Change Runtime Data
    async def _async_change_runtime_data(
        self,
        operation: Literal["set", "append", "del"],
        key: str,
        value: Any,
        *,
        emit: bool = True,
    ):
        futures = []
        handlers = self._handlers["runtime_data"]

        match operation:
            case "set":
                self._runtime_data.set(key, value)
                value = self._runtime_data[key]
            case "append":
                self._runtime_data.append(key, value)
                value = self._runtime_data[key]
            case "del":
                if self._runtime_data.get(key, None):
                    del self._runtime_data[key]
                    value = None
                else:
                    return
        if emit:
            if key in handlers:
                futures.append(
                    self.async_emit(
                        key,
                        value,
                        trigger_type="runtime_data",
                    )
                )

            if futures:
                await asyncio.gather(*futures, return_exceptions=self._skip_exceptions)

    async def async_set_runtime_data(
        self,
        key: str,
        value: Any,
        *,
        emit: bool = True,
    ):
        return await self._async_change_runtime_data("set", key, value, emit=emit)

    async def async_append_runtime_data(
        self,
        key: str,
        value: Any,
        *,
        emit: bool = True,
    ):
        return await self._async_change_runtime_data("append", key, value, emit=emit)

    async def async_del_runtime_data(
        self,
        key: str,
        *,
        emit: bool = True,
    ):
        return await self._async_change_runtime_data("del", key, None, emit=emit)

    # Start
    async def async_start(
        self,
        initial_value: Any = None,
        *,
        wait_for_result: bool = True,
        timeout: float | None = 10,
    ):
        if not self._started:
            self._started = True
            if self._status not in {TRIGGER_FLOW_STATUS_COMPLETED, TRIGGER_FLOW_STATUS_FAILED, TRIGGER_FLOW_STATUS_CANCELLED}:
                self._set_status(TRIGGER_FLOW_STATUS_RUNNING)
            await self._async_dispatch_signal(
                self._build_signal(
                    "START",
                    initial_value,
                    trigger_type="event",
                    source="start",
                )
            )
        if wait_for_result:
            return await self.async_get_result(timeout=timeout)

    # Pause / Continue
    async def async_pause_for(
        self,
        *,
        type: str = "pause",
        payload: Any = None,
        resume_event: str | None = None,
        interrupt_id: str | None = None,
    ):
        interrupt_id = interrupt_id if interrupt_id is not None else uuid.uuid4().hex
        interrupts = self._get_interrupts().copy()
        interrupt = {
            "id": interrupt_id,
            "type": type,
            "payload": payload,
            "resume_event": resume_event,
            "status": "waiting",
        }
        interrupts[interrupt_id] = interrupt
        self._system_runtime_data.set("interrupts", interrupts)
        self._set_status(TRIGGER_FLOW_STATUS_WAITING)
        await self.async_put_into_stream(
            {
                "type": "interrupt",
                "action": "pause",
                "execution_id": self.id,
                "interrupt": self._to_serializable_value(interrupt),
                "signal": self._serialize_signal(self.get_last_signal()),
            }
        )
        return TriggerFlowPauseSignal(interrupt)

    async def async_continue_with(
        self,
        interrupt_id: str,
        value: Any = None,
    ):
        interrupts = self._get_interrupts().copy()
        if interrupt_id not in interrupts:
            raise KeyError(f"Can not continue execution { self.id }, interrupt '{ interrupt_id }' not found.")
        interrupt = dict(interrupts[interrupt_id])
        if interrupt.get("status") != "waiting":
            raise ValueError(
                f"Can not continue execution { self.id }, interrupt '{ interrupt_id }' is not waiting."
            )
        interrupt["status"] = "resumed"
        interrupt["response"] = value
        interrupts[interrupt_id] = interrupt
        self._system_runtime_data.set("interrupts", interrupts)
        self._set_status(TRIGGER_FLOW_STATUS_RUNNING)
        await self.async_put_into_stream(
            {
                "type": "interrupt",
                "action": "resume",
                "execution_id": self.id,
                "interrupt": self._to_serializable_value(interrupt),
                "value": self._to_serializable_value(value),
            }
        )
        resume_event = interrupt.get("resume_event")
        if resume_event:
            await self._async_dispatch_signal(
                self._build_signal(
                    str(resume_event),
                    value,
                    trigger_type="event",
                    source="interrupt",
                    meta={"interrupt_id": interrupt_id},
                )
            )
        return interrupt

    # Runtime Stream
    async def async_put_into_stream(self, stream_item: Any):
        await self._runtime_stream_queue.put(stream_item)

    async def async_stop_stream(self):
        await self._runtime_stream_queue.put(RUNTIME_STREAM_STOP)

    async def _consume_runtime_stream(
        self,
        *,
        initial_value: Any,
        timeout: float | None,
    ):
        temp_execution_task = None
        try:
            if not self._started:
                temp_execution_task = asyncio.create_task(
                    self.async_start(
                        initial_value=initial_value,
                        wait_for_result=False,
                    )
                )
            while True:
                if timeout is not None:
                    try:
                        next_result = await asyncio.wait_for(
                            self._runtime_stream_queue.get(),
                            timeout=timeout,
                        )
                    except asyncio.TimeoutError:
                        warnings.warn(
                            f"Execution { self.id } runtime stream stopped because of timeout.\n"
                            f"Timeout seconds: { timeout }\n"
                            "You can use execution.get_async_runtime_stream(timeout=<int | None>) or execution.get_runtime_stream(timeout=<int | None>) to reset new timeout seconds or use None to wait forever."
                        )
                        break
                else:
                    next_result = await self._runtime_stream_queue.get()
                if next_result is not RUNTIME_STREAM_STOP:
                    yield next_result
                else:
                    break
        finally:
            if temp_execution_task:
                await temp_execution_task

    def get_async_runtime_stream(
        self,
        initial_value: Any = None,
        *,
        timeout: float | None = 10,
    ):
        if self._runtime_stream_consumer is None:
            self._runtime_stream_consumer = GeneratorConsumer(
                self._consume_runtime_stream(
                    initial_value=initial_value,
                    timeout=timeout,
                )
            )
        return self._runtime_stream_consumer.get_async_generator()

    def get_runtime_stream(
        self,
        initial_value: Any = None,
        *,
        timeout: float | None = 10,
    ):
        if self._runtime_stream_consumer is None:
            self._runtime_stream_consumer = GeneratorConsumer(
                self._consume_runtime_stream(
                    initial_value=initial_value,
                    timeout=timeout,
                )
            )
        return self._runtime_stream_consumer.get_generator()

    # Result
    def set_result(self, result: Any):
        self._system_runtime_data.set("result", result)
        result_ready = self._system_runtime_data.get("result_ready")
        if isinstance(result_ready, asyncio.Event):
            result_ready.set()
        self._set_status(TRIGGER_FLOW_STATUS_COMPLETED)

    async def async_get_result(self, *, timeout: float | None = None):
        if timeout is None:
            result_ready = self._system_runtime_data.get("result_ready")
            if isinstance(result_ready, asyncio.Event):
                await result_ready.wait()
            self._result = self._system_runtime_data.get("result")
            return self._result
        else:
            try:
                result_ready = self._system_runtime_data.get("result_ready")
                if isinstance(result_ready, asyncio.Event):
                    await asyncio.wait_for(result_ready.wait(), timeout=timeout)
                self._result = self._system_runtime_data.get("result")
                return self._result
            except asyncio.TimeoutError:
                warnings.warn(
                    f"Can not get the result of trigger flow { self.id } for it took too long and timeout.\n"
                    "You can check if you forget to use flow.set_result() to set a result for this trigger flow. Or you can set parameter 'timeout' to a bigger number to wait longer or to None to wait forever."
                    f"Timeout: { timeout }"
                )
                self._result = None
                return self._result



# Copyright 2023-2026 AgentEra(Agently.Tech)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .StreamingPrintExtension import StreamingPrintExtension
from .ToolExtension import ToolExtension
from .KeyWaiterExtension import KeyWaiterExtension
from .AutoFuncExtension import AutoFuncExtension
from .ConfigurePromptExtension import ConfigurePromptExtension
from .SessionExtension import SessionExtension
from .ChatSessionExtension import ChatSessionExtension



# Copyright 2023-2026 AgentEra(Agently.Tech)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from uuid import uuid4

from typing import TYPE_CHECKING, Sequence

from agently.core import BaseAgent, Session

if TYPE_CHECKING:
    from agently.core import Prompt
    from agently.core.ModelRequest import ModelResponseResult
    from agently.types.data import ChatMessage, ChatMessageDict
    from agently.utils import Settings
    from agently.types.plugins import SessionAnalysisHandler, SessionResizeHandler


class SessionExtension(BaseAgent):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.sessions: dict[str, Session] = {}
        self.activated_session: Session | None = None
        self.settings.setdefault("session.input_keys", None, inherit=True)
        self.settings.setdefault("session.reply_keys", None, inherit=True)

        self.__session_analysis_handler: "SessionAnalysisHandler | None" = None
        self.__session_resize_handlers: dict[str, "SessionResizeHandler"] = {}

        self.extension_handlers.append("request_prefixes", self._session_request_prefix)
        self.extension_handlers.append("finally", self._session_finally)

    def register_session_analysis_handler(self, handler: "SessionAnalysisHandler"):
        """
        Register analysis handler for session resize planning.

        Signature:
            `(full_context, context_window, memo, session_settings) -> str | None`.
        """
        self.__session_analysis_handler = handler
        for session in self.sessions.values():
            session.register_analysis_handler(handler)
        return self

    def register_session_resize_handler(self, strategy_name: str, handler: "SessionResizeHandler"):
        """
        Register resize handler for a strategy.

        Signature:
            `(full_context, context_window, memo, session_settings) -> (new_full_context, new_context_window, new_memo)`.
        """
        if not isinstance(strategy_name, str) or strategy_name.strip() == "":
            raise ValueError("strategy_name must be a non-empty string.")
        self.__session_resize_handlers[strategy_name] = handler
        for session in self.sessions.values():
            session.register_resize_handler(strategy_name, handler)
        return self

    def __bind_session_resize_pipeline(self, session: Session):
        session.register_analysis_handler(self.__session_analysis_handler)
        for strategy_name, handler in self.__session_resize_handlers.items():
            session.register_resize_handler(strategy_name, handler)

    def _refill_agent_chat_history_with_session(self):
        if self.activated_session is None:
            return self
        if "chat_history" in self.agent_prompt:
            del self.agent_prompt["chat_history"]
        self.agent_prompt.set("chat_history", self.activated_session.context_window)
        return self

    def activate_session(self, *, session_id: str | None = None):
        if session_id is not None and session_id in self.sessions:
            self.activated_session = self.sessions[session_id]
        else:
            if session_id is None:
                session_id = uuid4().hex
            self.activated_session = Session(
                id=session_id,
                auto_resize=True,
                settings=self.settings,
            )
            self.sessions[session_id] = self.activated_session

        self.__bind_session_resize_pipeline(self.activated_session)
        return self._refill_agent_chat_history_with_session()

    def deactivate_session(self):
        self.activated_session = None
        if "chat_history" in self.agent_prompt:
            del self.agent_prompt["chat_history"]
        self.agent_prompt.set("chat_history", [])
        return self

    def reset_chat_history(self):
        if self.activated_session is None:
            return super().reset_chat_history()
        self.activated_session.reset_chat_history()
        return self._refill_agent_chat_history_with_session()

    def set_chat_history(
        self,
        chat_history: "Sequence[ChatMessage | ChatMessageDict]",
    ):
        if self.activated_session is None:
            return super().set_chat_history(chat_history)
        self.activated_session.set_chat_history(chat_history)
        return self._refill_agent_chat_history_with_session()

    def add_chat_history(
        self,
        chat_history: "Sequence[ChatMessage | ChatMessageDict] | ChatMessage | ChatMessageDict",
    ):
        if self.activated_session is None:
            return super().add_chat_history(chat_history)
        self.activated_session.add_chat_history(chat_history)
        return self._refill_agent_chat_history_with_session()

    def clean_context_window(self):
        if self.activated_session is None:
            if "chat_history" in self.agent_prompt:
                del self.agent_prompt["chat_history"]
            return self
        self.activated_session.clean_context_window()
        return self._refill_agent_chat_history_with_session()

    async def _session_request_prefix(self, prompt: "Prompt", _settings: "Settings"):
        Session.apply_request_prefix(prompt, self.activated_session)

    async def _session_finally(self, result: "ModelResponseResult", settings: "Settings"):
        if self.activated_session is None:
            return

        user_content, assistant_content = await Session.resolve_finally_contents(
            result,
            settings,
            self.agent_prompt,
        )

        if user_content is not None and user_content != "":
            self.add_chat_history({"role": "user", "content": user_content})
        if assistant_content is not None and assistant_content != "":
            self.add_chat_history({"role": "assistant", "content": assistant_content})



import uuid

from dataclasses import dataclass, field
from typing import Any, Literal


TriggerFlowSignalType = Literal["event", "runtime_data", "flow_data"]


@dataclass(slots=True)
class TriggerFlowSignal:
    id: str
    trigger_event: str
    trigger_type: TriggerFlowSignalType
    value: Any = None
    layer_marks: list[str] = field(default_factory=list)
    source: str = "runtime"
    meta: dict[str, Any] = field(default_factory=dict)

    @classmethod
    def create(
        cls,
        *,
        trigger_event: str,
        trigger_type: TriggerFlowSignalType = "event",
        value: Any = None,
        layer_marks: list[str] | None = None,
        source: str = "runtime",
        meta: dict[str, Any] | None = None,
    ):
        return cls(
            id=uuid.uuid4().hex,
            trigger_event=trigger_event,
            trigger_type=trigger_type,
            value=value,
            layer_marks=layer_marks.copy() if layer_marks is not None else [],
            source=source,
            meta=meta.copy() if meta is not None else {},
        )

    def to_debug_dict(self):
        return {
            "SIGNAL_ID": self.id,
            "TYPE": self.trigger_type,
            "EVENT": self.trigger_event,
            "VALUE": self.value,
            "SOURCE": self.source,
            "META": self.meta,
        }

    def to_state_dict(self):
        return {
            "id": self.id,
            "trigger_event": self.trigger_event,
            "trigger_type": self.trigger_type,
            "value": self.value,
            "layer_marks": self.layer_marks.copy(),
            "source": self.source,
            "meta": self.meta.copy(),
        }



# Copyright 2023-2026 AgentEra(Agently.Tech)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class AVOID_COPY:
    __slots__ = ("id",)

    def __init__(self):
        import uuid

        self.id = uuid.uuid4().hex

    def __copy__(self):
        return self

    def __deepcopy__(self, memo):
        return self

    def __reduce__(self):
        return (self.__class__, (), {"id": self.id})


EMPTY = AVOID_COPY()

from .serializable import SerializableData, SerializableValue
from .prompt import (
    ChatMessage,
    ChatMessageDict,
    ChatMessageContent,
    TextMessageContent,
    PromptModel,
    PromptStandardSlot,
    ToolMeta,
)
from .request import (
    AgentlyRequestData,
    AgentlyRequestDataDict,
)

from .response import (
    AgentlyModelResult,
    AgentlyModelResponseEvent,
    AgentlyModelResponseMessage,
    AgentlyResponseGenerator,
    InstantStreamingContentType,
    NormalStreamingContentType,
    ResponseContentType,
    SpecificEvents,
    StreamingContentType,
    StreamingData,
)

from .event import (
    AgentlyEvent,
    AgentlySystemEvent,
    MessageLevel,
    EventStatus,
    EventMessageDict,
    EventMessage,
    EventHook,
)

from .tool import (
    ArgumentDesc,
    KwargsType,
    ReturnType,
    MCPConfig,
    MCPConfigs,
    ToolInfo,
)



# Copyright 2023-2026 AgentEra(Agently.Tech)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import uuid
from collections.abc import Mapping

from typing import Any, Callable, Literal, TYPE_CHECKING
from agently.types.data import AVOID_COPY

if TYPE_CHECKING:
    from agently.core import TriggerFlowExecution
    from agently.core.TriggerFlow.Signal import TriggerFlowSignal

from agently.utils import RuntimeData


class TriggerFlowBlockData:
    global_data = RuntimeData()

    def __init__(
        self,
        outer_block: "TriggerFlowBlockData | None" = None,
        data: dict[str, Any] | None = None,
    ):
        self.outer_block = outer_block
        self.data = data if data is not None else {}


_MISSING = object()


class _TriggerFlowDataNamespace:
    def __init__(
        self,
        *,
        getter: Callable[..., Any],
        setter: Callable[..., Any],
        appender: Callable[..., Any],
        deleter: Callable[..., Any],
        async_setter: Callable[..., Any],
        async_appender: Callable[..., Any],
        async_deleter: Callable[..., Any],
    ):
        self._getter = getter
        self._setter = setter
        self._appender = appender
        self._deleter = deleter
        self.async_set = async_setter
        self.async_append = async_appender
        self.async_del = async_deleter

    def get(
        self,
        key: Any | None = None,
        default: Any = None,
        *,
        inherit: bool = True,
    ):
        return self._getter(key, default, inherit=inherit)

    def set(
        self,
        key: str,
        value: Any,
        *,
        emit: bool = True,
    ):
        return self._setter(key, value, emit=emit)

    def append(
        self,
        key: str,
        value: Any,
        *,
        emit: bool = True,
    ):
        return self._appender(key, value, emit=emit)

    def delete(
        self,
        key: str,
        *,
        emit: bool = True,
    ):
        return self._deleter(key, emit=emit)

    def to_dict(self, *, inherit: bool = True):
        data = self.get(None, {}, inherit=inherit)
        return data if isinstance(data, dict) else {}

    def keys(self):
        return self.to_dict().keys()

    def values(self):
        return self.to_dict().values()

    def items(self):
        return self.to_dict().items()

    def __getitem__(self, key: Any):
        value = self.get(key, _MISSING)
        if value is _MISSING:
            raise KeyError(key)
        return value

    def __contains__(self, key: Any):
        return self.get(key, _MISSING) is not _MISSING

    def __iter__(self):
        return iter(self.to_dict())

    def __len__(self):
        return len(self.to_dict())


class _TriggerFlowResourcesView(Mapping[str, Any]):
    def __init__(self, execution: "TriggerFlowExecution"):
        self._execution = execution

    def to_dict(self):
        return self._execution.get_runtime_resources()

    def __getitem__(self, key: str):
        value = self._execution.get_runtime_resource(key, _MISSING)
        if value is _MISSING:
            raise KeyError(key)
        return value

    def __iter__(self):
        return iter(self.to_dict())

    def __len__(self):
        return len(self.to_dict())

    def get(self, key: str, default: Any = None):
        return self._execution.get_runtime_resource(key, default)


class _TriggerFlowSignalInfo:
    def __init__(
        self,
        *,
        trigger_event: str,
        trigger_type: Literal["event", "runtime_data", "flow_data", "collect"],
        value: Any,
        signal: "TriggerFlowSignal | None",
        signal_id: str | None,
        signal_source: str | None,
        signal_meta: dict[str, Any],
    ):
        self.trigger_event = trigger_event
        self.trigger_type = trigger_type
        self.value = value
        self.signal = signal
        self.signal_id = signal_id
        self.signal_source = signal_source
        self.signal_meta = signal_meta

    @property
    def event(self):
        return self.trigger_event

    @property
    def type(self):
        return self.trigger_type

    @property
    def source(self):
        return self.signal_source

    @property
    def meta(self):
        return self.signal_meta

    def to_dict(self):
        return {
            "trigger_event": self.trigger_event,
            "trigger_type": self.trigger_type,
            "value": self.value,
            "signal_id": self.signal_id,
            "signal_source": self.signal_source,
            "signal_meta": self.signal_meta.copy(),
        }


class TriggerFlowRuntimeData:
    def __init__(
        self,
        *,
        trigger_event: str,
        trigger_type: Literal["event", "runtime_data", "flow_data", "collect"],
        value: Any,
        execution: "TriggerFlowExecution",
        _layer_marks: list[str] | None = None,
        signal: "TriggerFlowSignal | None" = None,
    ):
        self.trigger_event = trigger_event
        self.trigger_type: Literal["event", "runtime_data", "flow_data", "collect"] = trigger_type
        self.event = trigger_event
        self.type = trigger_type
        self.value = value
        self.execution = execution
        self.execution_id = execution.id
        self._layer_marks = _layer_marks if _layer_marks is not None else []
        self.settings = execution.settings
        self.signal = signal
        self.signal_id = signal.id if signal is not None else None
        self.signal_source = signal.source if signal is not None else None
        self.signal_meta = signal.meta.copy() if signal is not None else {}
        self.signal_info = _TriggerFlowSignalInfo(
            trigger_event=trigger_event,
            trigger_type=trigger_type,
            value=value,
            signal=signal,
            signal_id=self.signal_id,
            signal_source=self.signal_source,
            signal_meta=self.signal_meta.copy(),
        )

        self.get_flow_data = execution.get_flow_data
        self.set_flow_data = execution.set_flow_data
        self.append_flow_data = execution.append_flow_data
        self.del_flow_data = execution.del_flow_data
        self.async_set_flow_data = execution.async_set_flow_data
        self.async_append_flow_data = execution.async_append_flow_data
        self.async_del_flow_data = execution.async_del_flow_data

        self.get_runtime_data = execution.get_runtime_data
        self.set_runtime_data = execution.set_runtime_data
        self.append_runtime_data = execution.append_runtime_data
        self.del_runtime_data = execution.del_runtime_data
        self.async_set_runtime_data = execution.async_set_runtime_data
        self.async_append_runtime_data = execution.async_append_runtime_data
        self.async_del_runtime_data = execution.async_del_runtime_data
        self.state = _TriggerFlowDataNamespace(
            getter=self.get_runtime_data,
            setter=self.set_runtime_data,
            appender=self.append_runtime_data,
            deleter=self.del_runtime_data,
            async_setter=self.async_set_runtime_data,
            async_appender=self.async_append_runtime_data,
            async_deleter=self.async_del_runtime_data,
        )
        self.flow_state = _TriggerFlowDataNamespace(
            getter=self.get_flow_data,
            setter=self.set_flow_data,
            appender=self.append_flow_data,
            deleter=self.del_flow_data,
            async_setter=self.async_set_flow_data,
            async_appender=self.async_append_flow_data,
            async_deleter=self.async_del_flow_data,
        )
        self.resources = _TriggerFlowResourcesView(execution)

        self.get_resource = execution.get_runtime_resource
        self.require_resource = execution.require_runtime_resource
        self.set_resource = execution.set_runtime_resource
        self.del_resource = execution.del_runtime_resource

        self.emit = execution.emit
        self.async_emit = execution.async_emit

        self.put = execution.put_into_stream
        self.async_put = execution.async_put_into_stream
        self.put_into_stream = execution.put_into_stream
        self.async_put_into_stream = execution.async_put_into_stream
        self.stop_stream = execution.stop_stream
        self.async_stop_stream = execution.async_stop_stream

        self.pause_for = execution.pause_for
        self.async_pause_for = execution.async_pause_for
        self.continue_with = execution.continue_with
        self.async_continue_with = execution.async_continue_with
        self.get_status = execution.get_status
        self.is_waiting = execution.is_waiting
        self.get_interrupt = execution.get_interrupt
        self.get_pending_interrupts = execution.get_pending_interrupts

        self.set_result = execution.set_result
        self.get_result = execution.get_result
        self.get_last_signal = execution.get_last_signal

        self._system_runtime_data = execution._system_runtime_data

    @property
    def upper_layer_mark(self):
        return self._layer_marks[-2] if len(self._layer_marks) > 1 else None

    @property
    def layer_mark(self):
        return self._layer_marks[-1] if len(self._layer_marks) > 0 else None

    def layer_in(self):
        self._layer_marks.append(uuid.uuid4().hex)

    def layer_out(self):
        self._layer_marks = self._layer_marks[:-1] if len(self._layer_marks) > 0 else []


TriggerFlowEventData = TriggerFlowRuntimeData
TriggerFlowHandler = Callable[[TriggerFlowRuntimeData], Any]
TriggerFlowHandlers = dict[str, dict[str, TriggerFlowHandler]]
TriggerFlowAllHandlers = dict[Literal["event", "flow_data", "runtime_data"], TriggerFlowHandlers]

RUNTIME_STREAM_STOP = AVOID_COPY()



# Copyright 2023-2026 AgentEra(Agently.Tech)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import uuid
import copy
from asyncio import Event, Semaphore

from typing import Callable, Any, Literal, TYPE_CHECKING, overload, cast
from typing_extensions import Self


if TYPE_CHECKING:
    from ..BluePrint import TriggerFlowBluePrint
    from agently.types.trigger_flow import TriggerFlowHandler, TriggerFlowRuntimeData

from ..Chunk import TriggerFlowChunk
from agently.types.data import EMPTY
from agently.types.trigger_flow import TriggerFlowBlockData


class TriggerFlowBaseProcess:

    def __init__(
        self,
        *,
        flow_chunk,
        trigger_event: str,
        blue_print: "TriggerFlowBluePrint",
        block_data: "TriggerFlowBlockData",
        trigger_type: Literal["event", "runtime_data", "flow_data"] = "event",
        definition_signals: list[dict[str, Any]] | None = None,
        definition_group_id: str | None = None,
        definition_group_kind: str | None = None,
        **options,
    ):
        self._flow_chunk = flow_chunk
        self.trigger_event = trigger_event
        self.trigger_type: Literal["event", "runtime_data", "flow_data"] = trigger_type
        self._blue_print = blue_print
        self._block_data = block_data
        self._options = options
        self._definition_signals = copy.deepcopy(definition_signals) if definition_signals is not None else []
        self._definition_group_id = definition_group_id
        self._definition_group_kind = definition_group_kind

    def _new(
        self,
        trigger_event: str,
        blue_print: "TriggerFlowBluePrint",
        block_data: "TriggerFlowBlockData",
        trigger_type: Literal["event", "runtime_data", "flow_data"] = "event",
        definition_signals: list[dict[str, Any]] | None = None,
        definition_group_id: str | None = None,
        definition_group_kind: str | None = None,
        **options,
    ):
        return type(self)(
            flow_chunk=self._flow_chunk,
            trigger_event=trigger_event,
            trigger_type=trigger_type,
            blue_print=blue_print,
            block_data=block_data,
            definition_signals=definition_signals if definition_signals is not None else self._definition_signals,
            definition_group_id=(
                definition_group_id if definition_group_id is not None else self._definition_group_id
            ),
            definition_group_kind=(
                definition_group_kind if definition_group_kind is not None else self._definition_group_kind
            ),
            **options,
        )

    def _layer_key(self, data: "TriggerFlowRuntimeData"):
        return ".".join(data._layer_marks) if data._layer_marks else "__root__"

    def _root_block_data(self):
        current = self._block_data
        while current.outer_block is not None:
            current = current.outer_block
        return current

    def _event_signal(self, trigger_event: str, *, role: str | None = None):
        return self._blue_print.make_signal("event", trigger_event, role=role)

    def _resolve_definition_signal(
        self,
        trigger_type: Literal["event", "runtime_data", "flow_data"],
        trigger_event: str | TriggerFlowChunk,
    ):
        if isinstance(trigger_event, TriggerFlowChunk):
            if trigger_type != "event":
                raise TypeError("Can not use chunk as trigger event when trigger type is not 'event'.")
            return self._event_signal(trigger_event.trigger)
        if trigger_type == "event" and isinstance(trigger_event, str) and trigger_event in self._blue_print.chunks:
            return self._event_signal(self._blue_print.chunks[trigger_event].trigger)
        return self._blue_print.make_signal(trigger_type, str(trigger_event))

    @overload
    def when(
        self,
        trigger_or_triggers: str | TriggerFlowChunk,
        *,
        mode: Literal["and", "or", "simple_or"] = "and",
    ) -> Self: ...

    @overload
    def when(
        self,
        trigger_or_triggers: dict[
            Literal["event"],
            str | TriggerFlowChunk | list[str | TriggerFlowChunk],
        ],
        *,
        mode: Literal["and", "or", "simple_or"] = "and",
    ) -> Self: ...

    @overload
    def when(
        self,
        trigger_or_triggers: dict[
            Literal["runtime_data", "flow_data", "collect"],
            str | list[str],
        ],
        *,
        mode: Literal["and", "or", "simple_or"] = "and",
    ) -> Self: ...

    def when(
        self,
        trigger_or_triggers: (
            str
            | TriggerFlowChunk
            | dict[
                Literal["event"],
                str | TriggerFlowChunk | list[str | TriggerFlowChunk],
            ]
            | dict[
                Literal["runtime_data", "flow_data", "collect"],
                str | list[str],
            ]
        ),
        *,
        mode: Literal["and", "or", "simple_or"] = "and",
    ):
        if isinstance(trigger_or_triggers, TriggerFlowChunk):
            trigger_or_triggers = trigger_or_triggers.trigger
        if isinstance(trigger_or_triggers, str):
            if trigger_or_triggers in self._blue_print.chunks:
                trigger = self._blue_print.chunks[trigger_or_triggers].trigger
            else:
                trigger = trigger_or_triggers
            return self._new(
                trigger_event=trigger,
                trigger_type="event",
                blue_print=self._blue_print,
                block_data=TriggerFlowBlockData(
                    outer_block=None,
                ),
                definition_signals=[self._event_signal(trigger)],
                definition_group_id=None,
                definition_group_kind=None,
            )
        values: dict[Literal["event", "runtime_data", "flow_data", "collect"], dict[str, Any]] = {}
        definition_signals: list[dict[str, Any]] = []
        trigger_count = 0
        current_trigger_type = "event"
        current_trigger_event = ""
        for trigger_type, trigger_event_or_events in trigger_or_triggers.items():
            if trigger_type not in values:
                values[trigger_type] = {}
            if isinstance(trigger_event_or_events, TriggerFlowChunk):
                if trigger_type == "event":
                    trigger_event_or_events = trigger_event_or_events.trigger
                else:
                    raise TypeError("Can not use chunk as trigger event when trigger type is not 'event'.")
            if trigger_type == "collect":
                trigger_type = "event"
                if isinstance(trigger_event_or_events, str):
                    trigger_event_or_events = f"Collect-{ trigger_event_or_events }"
                else:
                    trigger_event_or_events = [f"Collect-{ trigger_event }" for trigger_event in trigger_event_or_events]
            if isinstance(trigger_event_or_events, str):
                values[trigger_type][trigger_event_or_events] = EMPTY
                definition_signals.append(self._resolve_definition_signal(trigger_type, trigger_event_or_events))
                current_trigger_type = trigger_type
                current_trigger_event = trigger_event_or_events
                trigger_count += 1
            else:
                for trigger_event in trigger_event_or_events:
                    if isinstance(trigger_event, TriggerFlowChunk):
                        trigger_event = trigger_event.trigger
                    values[trigger_type][trigger_event] = EMPTY
                    definition_signals.append(self._resolve_definition_signal(trigger_type, trigger_event))
                    current_trigger_type = trigger_type
                    current_trigger_event = trigger_event
                    trigger_count += 1

        if trigger_count == 1:
            return self._new(
                trigger_event=current_trigger_event,
                trigger_type=current_trigger_type,
                blue_print=self._blue_print,
                block_data=TriggerFlowBlockData(
                    outer_block=None,
                ),
                definition_signals=definition_signals,
                definition_group_id=None,
                definition_group_kind=None,
            )

        when_id = uuid.uuid4().hex
        when_trigger = f"When-{ when_id }"
        values_template = copy.deepcopy(values)

        async def wait_trigger(data: "TriggerFlowRuntimeData"):
            match mode:
                case "or" | "simple_or":
                    await data.async_emit(
                        when_trigger,
                        (
                            data.value
                            if mode == "simple_or"
                            else (data.trigger_type, data.trigger_event, data.value)
                        ),
                        _layer_marks=data._layer_marks.copy(),
                    )
                case "and":
                    state_key = f"when_states.{ when_id }.{ self._layer_key(data) }"
                    state = data._system_runtime_data.get(state_key)
                    if not isinstance(state, dict):
                        state = copy.deepcopy(values_template)
                    if data.trigger_type in state and data.trigger_event in state[data.trigger_type]:
                        state[data.trigger_type][data.trigger_event] = data.value
                    data._system_runtime_data.set(state_key, state)
                    for trigger_event_dict in state.values():
                        for event_value in trigger_event_dict.values():
                            if event_value is EMPTY:
                                return

                    await data.async_emit(
                        when_trigger,
                        state,
                        _layer_marks=data._layer_marks.copy(),
                    )
                    del data._system_runtime_data[state_key]

        for trigger_type, trigger_event_dict in values.items():
            for trigger_event in trigger_event_dict.keys():
                self._blue_print.add_handler(
                    trigger_type,  # type: ignore[arg-type]
                    trigger_event,
                    wait_trigger,
                )

        self._blue_print.definition.add_operator(
            id=f"when-{ when_id }",
            kind="signal_gate",
            name=f"when:{ when_id }",
            listen_signals=definition_signals,
            emit_signals=[self._event_signal(when_trigger, role="continuation")],
            options={"mode": mode},
        )

        return self._new(
            trigger_event=when_trigger,
            trigger_type="event",
            blue_print=self._blue_print,
            block_data=TriggerFlowBlockData(
                outer_block=None,
            ),
            definition_signals=[self._event_signal(when_trigger)],
            definition_group_id=None,
            definition_group_kind=None,
        )

    def to(
        self,
        chunk: "TriggerFlowChunk | TriggerFlowHandler | str | tuple[str, TriggerFlowHandler]",
        side_branch: bool = False,
        name: str | None = None,
    ):
        if isinstance(chunk, str):
            if chunk in self._blue_print.chunks:
                chunk = self._blue_print.chunks[chunk]
            else:
                raise NotImplementedError(f"Cannot find chunk named '{ chunk }'")
        elif isinstance(chunk, tuple):
            chunk_name = chunk[0]
            chunk_func = chunk[1]
            chunk = self._flow_chunk(chunk_name)(chunk_func)
        else:
            if callable(chunk):
                chunk = self._flow_chunk(chunk) if name is None else self._flow_chunk(name)(chunk)
        assert isinstance(chunk, TriggerFlowChunk)
        self._blue_print.add_handler(
            self.trigger_type,
            self.trigger_event,
            chunk.async_call,
        )
        self._blue_print.attach_chunk(
            chunk,
            self._definition_signals,
            group_id=self._definition_group_id,
            group_kind=self._definition_group_kind,
        )
        return self._new(
            trigger_event=chunk.trigger if not side_branch else self.trigger_event,
            trigger_type=self.trigger_type,
            blue_print=self._blue_print,
            block_data=self._block_data,
            definition_signals=[self._event_signal(chunk.trigger)] if not side_branch else self._definition_signals,
            definition_group_id=self._definition_group_id,
            definition_group_kind=self._definition_group_kind,
            **self._options,
        )

    def side_branch(
        self,
        chunk: "TriggerFlowChunk | TriggerFlowHandler",
        *,
        name: str | None = None,
    ):
        return self.to(
            chunk,
            side_branch=True,
            name=name,
        )

    def batch(
        self,
        *chunks: "TriggerFlowChunk | TriggerFlowHandler | tuple[str, TriggerFlowHandler]",
        side_branch: bool = False,
        concurrency: int | None = None,
    ):
        batch_id = uuid.uuid4().hex
        batch_trigger = f"Batch-{ batch_id }"
        results_template: dict[str, Any] = {}
        triggers_template: dict[str, bool] = {}
        trigger_to_chunk_name = {}
        branch_input_signals: list[dict[str, Any]] = []
        branch_output_signals: list[dict[str, Any]] = []
        result_keys: dict[str, str] = {}

        async def wait_all_chunks(data: "TriggerFlowRuntimeData"):
            if data.event not in trigger_to_chunk_name:
                return
            layer_key = self._layer_key(data)
            state_key = f"batch_states.{ batch_id }.{ layer_key }"
            state = data._system_runtime_data.get(state_key)
            if not isinstance(state, dict):
                state = {
                    "results": copy.deepcopy(results_template),
                    "triggers": triggers_template.copy(),
                }
            state["results"][trigger_to_chunk_name[data.event]] = data.value
            state["triggers"][data.event] = True
            data._system_runtime_data.set(state_key, state)
            for done in state["triggers"].values():
                if done is False:
                    return
            await data.async_emit(
                batch_trigger,
                state["results"],
                _layer_marks=data._layer_marks.copy(),
            )
            del data._system_runtime_data[state_key]

        for chunk in chunks:
            if isinstance(chunk, tuple):
                chunk_name = chunk[0]
                chunk_func = chunk[1]
                chunk = self._flow_chunk(chunk_name)(chunk_func)
            else:
                if callable(chunk):
                    chunk = self._flow_chunk(chunk)
            typed_chunk = cast(TriggerFlowChunk, chunk)
            triggers_template[typed_chunk.trigger] = False
            trigger_to_chunk_name[typed_chunk.trigger] = typed_chunk.name
            results_template[typed_chunk.name] = None

            branch_input_event = f"Batch-{ batch_id }-Input-{ typed_chunk.id }"
            branch_input_signal = self._event_signal(branch_input_event)
            branch_output_signal = self._event_signal(typed_chunk.trigger)
            branch_input_signals.append(branch_input_signal)
            branch_output_signals.append(branch_output_signal)
            result_keys[branch_output_signal["id"]] = typed_chunk.name

            if concurrency is None or concurrency <= 0:
                handler = typed_chunk.async_call
            else:

                def make_handler(bound_chunk: TriggerFlowChunk):
                    async def handler(data: "TriggerFlowRuntimeData"):
                        semaphore_key = f"batch_semaphores.{ batch_id }"
                        semaphore = data._system_runtime_data.get(semaphore_key, inherit=False)
                        if not isinstance(semaphore, Semaphore):
                            semaphore = Semaphore(concurrency)
                            data._system_runtime_data.set(semaphore_key, semaphore)
                        async with semaphore:
                            return await bound_chunk.async_call(data)

                    return handler

                handler = make_handler(typed_chunk)

            self._blue_print.add_handler(
                self.trigger_type,
                self.trigger_event,
                handler,
            )
            self._blue_print.add_event_handler(typed_chunk.trigger, wait_all_chunks)
            self._blue_print.attach_chunk(
                typed_chunk,
                [branch_input_signal],
                group_id=batch_id,
                group_kind="batch",
            )

        self._blue_print.definition.add_operator(
            id=f"batch-fanout-{ batch_id }",
            kind="batch_fanout",
            name=f"batch:{ batch_id }",
            listen_signals=self._definition_signals,
            emit_signals=branch_input_signals,
            options={"concurrency": concurrency},
            group_id=batch_id,
            group_kind="batch",
        )
        self._blue_print.definition.add_operator(
            id=f"batch-collect-{ batch_id }",
            kind="batch_collect",
            name=f"batch_collect:{ batch_id }",
            listen_signals=branch_output_signals,
            emit_signals=[self._event_signal(batch_trigger, role="continuation")],
            options={"result_keys": result_keys},
            group_id=batch_id,
            group_kind="batch",
        )

        return self._new(
            trigger_event=batch_trigger if not side_branch else self.trigger_event,
            blue_print=self._blue_print,
            block_data=self._block_data,
            definition_signals=[self._event_signal(batch_trigger)] if not side_branch else self._definition_signals,
            definition_group_id=self._definition_group_id,
            definition_group_kind=self._definition_group_kind,
            **self._options,
        )

    def collect(
        self,
        collection_name: str,
        branch_id: str | None = None,
        *,
        mode: Literal["filled_and_update", "filled_then_empty"] = "filled_and_update",
    ):
        root_block = self._root_block_data()
        if "_collect_configs" not in root_block.data:
            root_block.data["_collect_configs"] = {}
        collect_configs = root_block.data["_collect_configs"]
        if collection_name not in collect_configs:
            collect_configs[collection_name] = {
                "collect_id": uuid.uuid4().hex,
                "branch_ids": [],
                "definition_operator_ids": [],
            }
        collection_config = collect_configs[collection_name]

        branch_id = branch_id if branch_id is not None else uuid.uuid4().hex
        if branch_id not in collection_config["branch_ids"]:
            collection_config["branch_ids"].append(branch_id)

        collect_id = collection_config["collect_id"]
        collect_trigger = f"Collect-{ collection_name }"

        async def collect_branches(data: "TriggerFlowRuntimeData"):
            branch_ids = list(collection_config["branch_ids"])
            layer_key = self._layer_key(data)
            state_key = f"collect_states.{ collect_id }.{ layer_key }"
            state = data._system_runtime_data.get(state_key)
            if not isinstance(state, dict):
                state = {configured_branch_id: EMPTY for configured_branch_id in branch_ids}
            state[branch_id] = data.value
            data._system_runtime_data.set(state_key, state)

            for configured_branch_id in branch_ids:
                if state.get(configured_branch_id, EMPTY) is EMPTY:
                    return

            collected = {configured_branch_id: state[configured_branch_id] for configured_branch_id in branch_ids}
            await data.async_emit(
                collect_trigger,
                collected,
                _layer_marks=data._layer_marks.copy(),
            )
            if mode == "filled_then_empty":
                del data._system_runtime_data[state_key]

        self._blue_print.add_handler(
            self.trigger_type,
            self.trigger_event,
            collect_branches,
        )

        operator_id = f"collect-{ collect_id }-{ branch_id }"
        self._blue_print.definition.add_operator(
            id=operator_id,
            kind="collect_branch",
            name=f"collect:{ collection_name }",
            listen_signals=self._definition_signals,
            emit_signals=[self._event_signal(collect_trigger, role="continuation")],
            options={
                "collection_name": collection_name,
                "collect_id": collect_id,
                "branch_id": branch_id,
                "branch_ids": list(collection_config["branch_ids"]),
                "mode": mode,
            },
            group_id=collect_id,
            group_kind="collect",
        )
        collection_config["definition_operator_ids"].append(operator_id)
        for definition_operator_id in collection_config["definition_operator_ids"]:
            operator = self._blue_print.definition.get_operator(definition_operator_id)
            operator["options"]["branch_ids"] = list(collection_config["branch_ids"])

        return self._new(
            trigger_event=collect_trigger,
            blue_print=self._blue_print,
            block_data=self._block_data,
            definition_signals=[self._event_signal(collect_trigger)],
            definition_group_id=self._definition_group_id,
            definition_group_kind=self._definition_group_kind,
            **self._options,
        )

    def end(self):
        async def set_default_result(data: "TriggerFlowRuntimeData"):
            result = data._system_runtime_data.get("result")
            if result is EMPTY:
                data.set_result(data.value)
            else:
                result_ready = data._system_runtime_data.get("result_ready")
                if isinstance(result_ready, Event):
                    result_ready.set()

        self._blue_print.add_handler(
            self.trigger_type,
            self.trigger_event,
            set_default_result,
        )
        self._blue_print.definition.add_operator(
            id=f"result-{ uuid.uuid4().hex }",
            kind="result_sink",
            name="result",
            listen_signals=self._definition_signals,
        )
        return self._new(
            trigger_event=self.trigger_event,
            trigger_type=self.trigger_type,
            blue_print=self._blue_print,
            block_data=self._block_data,
            definition_signals=self._definition_signals,
            definition_group_id=self._definition_group_id,
            definition_group_kind=self._definition_group_kind,
            **self._options,
        )

    def ____(
        self,
        *args,
        log_info: bool = False,
        print_info: bool = False,
        show_value: bool = False,
        **kwargs,
    ):
        """
        Separator for chain expression.
        Do nothing but return self.
        """
        if log_info or print_info or show_value:

            async def runtime_output(data: "TriggerFlowRuntimeData"):
                from agently.base import async_system_message

                message = {}
                annotations = list(args) if args else []
                annotations.extend([f"{k}:{v}" for k, v in kwargs.items()])
                if annotations:
                    message["ANNOTATIONS"] = "\t".join(annotations)
                if show_value:
                    message["VALUE"] = data.value
                if log_info:
                    await async_system_message(
                        "TRIGGER_FLOW",
                        message,
                        data.settings,
                    )
                if print_info:
                    print(*message.values())

            self.side_branch(runtime_output)

        return self



# Copyright 2023-2026 AgentEra(Agently.Tech)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import datetime
import warnings
from typing import (
    Any,
    Literal,
    Mapping,
    Sequence,
    Union,
    get_origin,
    get_args,
    overload,
    TYPE_CHECKING,
    TypeVar,
)
from pydantic import BaseModel

if TYPE_CHECKING:
    from agently.types.data import SerializableValue, KwargsType
    from re import Pattern

T = TypeVar("T")

DEFAULT_PLACEHOLDER_PATTERN = re.compile(r"\$\{\s*([^}]+?)\s*\}")


class DataFormatter:
    @staticmethod
    def sanitize(value: Any, *, remain_type: bool = False) -> Any:
        from .RuntimeData import RuntimeData, RuntimeDataNamespace

        if isinstance(value, (str, int, float, bool)) or value is None:
            return value

        if isinstance(value, (datetime.datetime, datetime.date)):
            return value.isoformat()

        if issubclass(type(value), RuntimeData) or issubclass(type(value), RuntimeDataNamespace):
            return DataFormatter.sanitize(value.data, remain_type=remain_type)

        if isinstance(value, type):
            if issubclass(value, BaseModel):
                extracted_value = {}
                for name, field in value.model_fields.items():
                    annotation = field.annotation
                    if hasattr(field, "rebuild_annotation"):
                        try:
                            annotation = field.rebuild_annotation()
                        except Exception:
                            annotation = field.annotation
                    extracted_value.update(
                        {
                            name: (
                                (
                                    DataFormatter.sanitize(annotation, remain_type=remain_type),
                                    field.description,
                                )
                                if field.description
                                else (DataFormatter.sanitize(annotation, remain_type=remain_type),)
                            )
                        }
                    )
                return extracted_value
            else:
                if remain_type:
                    return value
                return value.__name__ if hasattr(value, "__name__") else str(value)
        if get_origin(value) is not None:
            if remain_type:
                return value
            original_text = get_origin(value)
            args = get_args(value)
            if original_text is list:
                return f"list[{DataFormatter.sanitize(args[0], remain_type=remain_type)}]"
            if original_text is dict:
                return f"dict[{DataFormatter.sanitize(args[0], remain_type=remain_type)}, {DataFormatter.sanitize(args[1], remain_type=remain_type)}]"
            if original_text is tuple:
                return f"tuple[{', '.join(DataFormatter.sanitize(a, remain_type=remain_type) for a in args)}]"
            if original_text is Union:
                return " | ".join(str(DataFormatter.sanitize(a, remain_type=remain_type)) for a in args)
            if original_text is Literal:
                return f"Literal[{ ', '.join(str(DataFormatter.sanitize(a, remain_type=remain_type)) for a in args) }]"
            if isinstance(value, type) and hasattr(value, "__name__"):
                return value.__name__
            return str(value)

        if isinstance(value, dict):
            return {str(k): DataFormatter.sanitize(v, remain_type=remain_type) for k, v in value.items()}
        if isinstance(value, list):
            return [DataFormatter.sanitize(v, remain_type=remain_type) for v in value]
        if isinstance(value, set):
            return {DataFormatter.sanitize(v, remain_type=remain_type) for v in value}
        if isinstance(value, tuple):
            return tuple(DataFormatter.sanitize(v, remain_type=remain_type) for v in value)

        return str(value)

    @overload
    @staticmethod
    def to_str_key_dict(
        value: Any,
        *,
        value_format: None = None,
        default_key: str,
        default_value: T = None,
        inconvertible_warning: bool = False,
    ) -> dict[str, Any | T]: ...

    @overload
    @staticmethod
    def to_str_key_dict(
        value: Any,
        *,
        value_format: Literal["serializable"],
        default_key: str,
        default_value: T = None,
        inconvertible_warning: bool = False,
    ) -> dict[str, "SerializableValue" | T]: ...

    @overload
    @staticmethod
    def to_str_key_dict(
        value: Any,
        *,
        value_format: Literal["str"],
        default_key: str,
        default_value: T = None,
        inconvertible_warning: bool = False,
    ) -> dict[str, str | T]: ...

    @overload
    @staticmethod
    def to_str_key_dict(
        value: Any,
        *,
        value_format: None = None,
        default_key: None,
        default_value: T = None,
        inconvertible_warning: bool = False,
    ) -> dict[str, Any] | T: ...

    @overload
    @staticmethod
    def to_str_key_dict(
        value: Any,
        *,
        value_format: Literal["serializable"],
        default_key: None,
        default_value: T = None,
        inconvertible_warning: bool = False,
    ) -> dict[str, "SerializableValue"] | T: ...

    @overload
    @staticmethod
    def to_str_key_dict(
        value: Any,
        *,
        value_format: Literal["str"],
        default_key: None,
        default_value: T = None,
        inconvertible_warning: bool = False,
    ) -> dict[str, str | T] | T: ...

    @overload
    @staticmethod
    def to_str_key_dict(
        value: Any,
        *,
        value_format: Literal['serializable', 'str'] | None = None,
        default_key: str | None = None,
        default_value: T = None,
        inconvertible_warning: bool = False,
    ) -> dict[str, Any] | T: ...

    @staticmethod
    def to_str_key_dict(
        value: Any,
        *,
        value_format: Literal["serializable", "str"] | None = None,
        default_key: str | None = None,
        default_value: T = None,
        inconvertible_warning: bool = False,
    ) -> dict[str, Any] | T:
        if isinstance(value, Mapping):
            if value_format is None:
                return {str(DataFormatter.sanitize(k)): v for k, v in value.items()}
            elif value_format == "serializable":
                return {str(DataFormatter.sanitize(k)): DataFormatter.sanitize(v) for k, v in value.items()}
            elif value_format == "str":
                return {str(DataFormatter.sanitize(k)): str(DataFormatter.sanitize(v)) for k, v in value.items()}
        else:
            if inconvertible_warning:
                warnings.warn(
                    f"Error: Non-dictionary value cannot be convert to a string key dictionary.\n"
                    f"Value: { value }\n"
                    "Tips: You can provide parameter 'default_key' to allow DataFormatter.to_str_key_dict() convert non-dictionary value to a dictionary { default_key: value } automatically."
                )
            # restructure value to {default_key: value}
            if default_key is not None:
                return DataFormatter.to_str_key_dict(
                    {default_key: value},
                    value_format=value_format,
                    default_value=default_value,
                )
            return default_value

    @staticmethod
    def to_str(value: Any) -> str:
        return str(DataFormatter.sanitize(value))

    @staticmethod
    def from_schema_to_kwargs_format(input_schema: dict[str, Any] | None) -> "KwargsType | None":
        if input_schema and len(input_schema.keys()) > 0:
            if "type" not in input_schema:
                raise KeyError(f"Cannot find key 'type' in input schema: { input_schema }")
            if input_schema["type"] != "object":
                raise TypeError(f"Input schema type is not 'object' but: { input_schema['type'] }")

            kwargs_format: "KwargsType" = {}

            if "properties" in input_schema and input_schema["properties"]:
                properties = input_schema["properties"]
                for kwarg_name, kwarg_schema in properties.items():
                    kwarg_type = kwarg_schema.pop("type", Any)
                    kwarg_schema.pop("title", None)
                    kwarg_desc = ";".join([f"{k}: {v}" for k, v in kwarg_schema.items()]) if kwarg_schema else ""
                    kwargs_format[kwarg_name] = (kwarg_type, kwarg_desc)

            if "additionalProperties" in input_schema:
                additional_properties = input_schema["additionalProperties"]
                if additional_properties is False:
                    pass
                elif additional_properties is True or additional_properties is None:
                    kwargs_format["<*>"] = (Any, "")
                elif isinstance(additional_properties, dict):
                    additional_type = additional_properties.pop("type", Any)
                    additional_properties.pop("title", None)
                    additional_desc = (
                        ";".join([f"{k}: {v}" for k, v in additional_properties.items()])
                        if additional_properties
                        else ""
                    )
                    kwargs_format["<*>"] = (additional_type, additional_desc)
                else:
                    kwargs_format["<*>"] = (Any, "")

            return kwargs_format or None

        return None

    @staticmethod
    def substitute_placeholder(
        obj: T,
        variable_mappings: dict[str, Any],
        *,
        placeholder_pattern: "Pattern | None" = None,
    ) -> T | Any:
        if placeholder_pattern is None:
            placeholder_pattern = DEFAULT_PLACEHOLDER_PATTERN

        if not isinstance(variable_mappings, dict):
            raise TypeError(f"Variable mappings require a dictionary but got: { variable_mappings }")

        if isinstance(obj, str):
            full_match = placeholder_pattern.fullmatch(obj)
            if full_match:
                key = full_match.group(1).strip()
                return variable_mappings.get(key, obj)
            else:

                def replacer(match):
                    key = match.group(1).strip()
                    return str(variable_mappings.get(key, match.group(0)))

                return placeholder_pattern.sub(replacer, obj)

        if isinstance(obj, Mapping):
            return {
                DataFormatter.substitute_placeholder(
                    key,
                    variable_mappings,
                    placeholder_pattern=placeholder_pattern,
                ): DataFormatter.substitute_placeholder(
                    value,
                    variable_mappings,
                    placeholder_pattern=placeholder_pattern,
                )
                for key, value in obj.items()
            }

        if isinstance(obj, Sequence) and not isinstance(obj, (str, bytes, bytearray)):
            if isinstance(obj, tuple):
                return tuple(
                    DataFormatter.substitute_placeholder(
                        value,
                        variable_mappings,
                        placeholder_pattern=placeholder_pattern,
                    )
                    for value in obj
                )
            else:
                return [
                    DataFormatter.substitute_placeholder(
                        value,
                        variable_mappings,
                        placeholder_pattern=placeholder_pattern,
                    )
                    for value in obj
                ]

        if isinstance(obj, set):
            return {
                DataFormatter.substitute_placeholder(
                    value,
                    variable_mappings,
                    placeholder_pattern=placeholder_pattern,
                )
                for value in obj
            }

        return obj



# Copyright 2023-2026 AgentEra(Agently.Tech)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
from copy import deepcopy
from typing import Any, Literal, Sequence, Mapping, cast, Iterator, TypeVar
from pathlib import Path

import json
import yaml
import toml

from agently.types.data import SerializableValue
from .DataFormatter import DataFormatter

T = TypeVar("T")


class DictRef:
    def __init__(self, container: dict[Any, Any], key: Any = None):
        self.container = container
        self.key = key

    def get(self) -> Any:
        if self.key is not None:
            return self.container[self.key]
        return self.container

    def set(self, value: Any):
        if self.key is not None:
            self.container[self.key] = value
        else:
            if isinstance(self.container, dict) and isinstance(value, dict):
                self.container.clear()
                self.container.update(value)
            else:
                raise TypeError("Setting root container to non-dict is not supported.")

    def update(self, new: dict[Any, Any]):
        ref = self.get()
        if isinstance(ref, dict):
            return ref.update(new)
        else:
            raise TypeError(f"Key '{ self.key }' from container { self.container } is not a dictionary.")

    def move_in(self, key: Any):
        return DictRef(self.get(), key)


class RuntimeData:
    instance_counter = 0

    def __init__(
        self,
        data: dict[Any, Any] | None = None,
        *,
        name: str | None = None,
        parent: "RuntimeData | None" = None,
    ):
        self._data = data if data is not None else {}
        if name is None:
            self.name = f"runtime_data_{ RuntimeData.instance_counter }"
            RuntimeData.instance_counter += 1
        else:
            self.name = name
        self.parent = parent

    def __repr__(self) -> str:
        return f"RuntimeData(name={ self.name }, data={ str(self.data) })"

    def __eq__(self, equal_target: Any) -> bool:
        return self.data == equal_target

    def __iter__(self) -> Iterator[Any]:
        """Make RuntimeData iterable like a standard dict"""
        return iter(self.data)

    def __len__(self) -> int:
        """Return the number of items in the RuntimeData"""
        return len(self.data)

    @property
    def data(self) -> dict[Any, Any]:
        return cast(dict[Any, Any], self.get(default={}))

    def _copy(self, origin: Any) -> Any:
        try:
            if isinstance(origin, dict):
                result = {}
                for key, value in origin.items():
                    result[key] = self._copy(value)
                return result
            if isinstance(origin, list):
                return [self._copy(item) for item in origin]
            if isinstance(origin, set):
                return {self._copy(item) for item in origin}
            if isinstance(origin, tuple):
                return tuple(self._copy(item) for item in origin)
            if hasattr(origin, "__class__") and hasattr(origin, "__module__") and origin.__module__ != "builtins":
                return origin
            return deepcopy(origin)
        except:
            return origin

    def _merge_view(self, child_data: dict[Any, Any], parent_data: dict[Any, Any]) -> dict[Any, Any]:
        result = self._copy(parent_data)
        for key, value in child_data.items():
            if key not in parent_data:
                result.update({key: self._copy(value)})
            elif isinstance(value, dict) and isinstance(parent_data[key], dict):
                result.update({key: self._merge_view(value, parent_data[key])})
            elif isinstance(value, list):
                if isinstance(parent_data[key], (list, set, tuple)):
                    for item in parent_data[key]:
                        if item not in value:
                            value.append(self._copy(item))
                elif parent_data[key] not in value:
                    value.append(self._copy(parent_data[key]))
                result.update({key: self._copy(value)})
            elif isinstance(value, set):
                if isinstance(parent_data[key], (list, set, tuple)):
                    for item in parent_data[key]:
                        if item not in value:
                            value.add(self._copy(item))
                else:
                    value.add(self._copy(parent_data[key]))
                result.update({key: self._copy(value)})
            else:
                result.update({key: self._copy(value)})
        return result

    def _get_inherited_view(self, runtime_data: "RuntimeData", result: dict[Any, Any] | None = None) -> dict[Any, Any]:
        if result is None:
            result = {}
        result = self._merge_view(result, runtime_data.get(default={}, inherit=False))
        if runtime_data.parent is not None:
            return self._get_inherited_view(runtime_data.parent, result)
        return result

    def _get_item_by_dot_path(self, dot_path: str, inherit: bool = True):
        current = self._get_inherited_view(self, {}) if inherit else self._copy(self._data)
        path_list = dot_path.split(".")
        for path in path_list:
            if path in current:
                current = current[path]
            else:
                return None
        return current

    def __getitem__(self, key: Any = None) -> Any:
        if isinstance(key, str) and "." in key:
            return self._get_item_by_dot_path(key)
        elif key is None:
            return self._data
        else:
            # Return None for missing keys, don't use data property to avoid inheritance
            return self.data.get(key)

    def get(
        self,
        key: Any | None = None,
        default: T = None,
        inherit: bool = True,
    ) -> Any | T:
        if key is None:
            if inherit:
                return self._get_inherited_view(self, {})
            else:
                return self._copy(self._data)

        # Handle key lookup
        if inherit:
            data = self.data
        else:
            data = self._data

        if isinstance(key, str) and "." in key:
            # Handle dot path
            current = data
            path_list = key.split(".")
            try:
                for path in path_list:
                    if isinstance(current, dict) and path in current:
                        current = current[path]
                    else:
                        return default
                return current
            except:
                return default
        else:
            # Handle simple key
            if isinstance(data, dict):
                # Use get with a sentinel to distinguish None values from missing keys
                sentinel = object()
                result = data.get(key, sentinel)
                if result is sentinel:
                    return default
                return result
            return default

    def keys(self):
        return self.data.keys()

    def values(self):
        return self.data.values()

    def items(self):
        return self.data.items()

    def pop(self, key: Any, default: Any = None) -> Any:
        if isinstance(key, str) and "." in key:
            val = self.get(key, default=None, inherit=False)
            if val is None:
                return default
            del self[key]
            return val
        else:
            if key in self._data:
                return self._data.pop(key)
            return default

    def clear(self):
        return self._data.clear()

    def __contains__(self, key: Any) -> bool:
        return key in self.data

    def _set_item(self, ref: DictRef, value: Any):
        if isinstance(ref.get(), dict) and isinstance(value, Mapping):
            for key, item_value in value.items():
                if key not in ref.get():
                    ref.get()[key] = self._copy(item_value)
                else:
                    self._set_item(ref.move_in(key), item_value)
            return
        if isinstance(ref.get(), list):
            if not isinstance(value, str) and isinstance(value, Sequence):
                for item in value:
                    if item not in ref.get():
                        ref.get().append(self._copy(item))
            elif value not in ref.get():
                ref.get().append(self._copy(value))
            return
        if isinstance(ref.get(), set):
            current_set = ref.get()
            from collections.abc import Iterable

            if not isinstance(value, (str, bytes)) and isinstance(value, Iterable):
                for item in value:
                    if item not in current_set:
                        current_set.add(self._copy(item))
            else:
                if value not in current_set:
                    current_set.add(self._copy(value))
            return
        else:
            existing = ref.get()
            if existing is None:
                if isinstance(value, list):
                    ref.set([self._copy(item) for item in value])
                else:
                    ref.set(self._copy(value))
            elif isinstance(existing, list):
                if isinstance(value, list):
                    for item in value:
                        if item not in existing:
                            existing.append(self._copy(item))
                else:
                    if value not in existing:
                        existing.append(self._copy(value))
            else:
                ref.set(self._copy(value))

    def _set_item_by_dot_path(self, dot_path: str, value: Any, *, cover: bool = False):
        current = DictRef(self._data)
        path_list = dot_path.split(".")
        walked_path = ""
        for path in path_list:
            if not isinstance(current.get(), dict):
                raise TypeError(
                    f"Can not set value to path '{ dot_path }' "
                    + f"because '{ walked_path[:-1] }' is not a dictionary."
                )
            walked_path += f"{ path }."
            if path not in current.get():
                current.update({path: {}})
            current = current.move_in(path)
        if cover:
            current.set(self._copy(value))
        else:
            self._set_item(current, value)

    def __setitem__(self, key: Any, value: Any):
        if isinstance(key, str) and "." in key:
            return self._set_item_by_dot_path(key, value)
        else:
            # For direct key assignment, use the merge behavior
            if key in self._data:
                existing = self._data[key]
                ref = DictRef(self._data, key)
                self._set_item(ref, value)
            else:
                self._data[key] = self._copy(value)

    def set(self, key: Any, value: Any):
        return self.__setitem__(key, value)

    def setdefault(self, key: Any, value: Any, *, inherit: bool = True):
        if self.get(key, inherit=inherit) is None:
            self.set(key, value)
        return self.get(key, inherit=inherit)

    def update(self, new: dict[Any, Any]):
        for key, value in new.items():
            self.set(key, value)

    def load(
        self,
        data_type: Literal[
            "json_file",
            "yaml_file",
            "toml_file",
            "json",
            "yaml",
            "toml",
        ],
        value: str,
    ):
        data = None
        if data_type.endswith("_file"):
            with open(value, "r", encoding="utf-8") as file:
                match data_type:
                    case "json_file":
                        data = json.load(file)
                    case "yaml_file":
                        data = yaml.safe_load(file)
                    case "toml_file":
                        data = toml.load(file)
                    case _:
                        raise TypeError(f"[Agently RuntimeData] Can not load type: '{ data_type }'")
        else:
            match data_type:
                case "json":
                    data = json.loads(value)
                case "yaml":
                    data = yaml.safe_load(value)
                case "toml":
                    data = toml.loads(value)
        if data and isinstance(data, dict):
            self.update(data)
        else:
            raise TypeError(
                f"[Agently RuntimeData] Can not load parsed data, expect dictionary type, got: { type(data) }"
            )

    def _get_serializable_data(self, data: Any) -> SerializableValue:
        if isinstance(data, (str, int, float, bool)) or data is None:
            return data
        elif isinstance(data, (datetime.datetime, datetime.date)):
            return data.isoformat()
        elif isinstance(data, RuntimeData):
            return self._get_serializable_data(data.data)
        elif isinstance(data, Path):
            return str(data)
        elif isinstance(data, dict):
            return {str(k): self._get_serializable_data(v) for k, v in data.items()}
        elif isinstance(data, (list, tuple, set)):
            return [self._get_serializable_data(v) for v in data]
        elif hasattr(data, "__dict__") and vars(data) != {}:
            return self._get_serializable_data(vars(data))
        else:
            return str(data)

    def dump(self, data_type: Literal["json", "yaml", "toml"]) -> str:
        serializable_data = self._get_serializable_data(self.data)
        if not isinstance(serializable_data, dict):
            raise TypeError("Can not dump not-dictionary runtime data.")
        match data_type:
            case "json":
                return json.dumps(serializable_data)
            case "yaml":
                return yaml.safe_dump(serializable_data)
            case "toml":
                return toml.dumps(DataFormatter.to_str_key_dict(serializable_data, default_key="data"))

    def __delitem__(self, key: Any):
        if isinstance(key, str) and "." in key:
            path_list = key.split(".")
            current = DictRef(self._data)
            for path in path_list[:-1]:
                cur = current.get()
                if not isinstance(cur, dict):
                    return
                if path in cur:
                    current = current.move_in(path)
                else:
                    return
            last_key = path_list[-1]
            cur = current.get()
            if isinstance(cur, dict) and last_key in cur:
                del cur[last_key]
        else:
            if key in self._data:
                del self._data[key]

    def append(self, key: Any, value: Any):
        if isinstance(key, str) and "." in key:
            current = self._get_item_by_dot_path(key, inherit=False)
        else:
            current = self._data.get(key)

        if current is None:
            new_value = [self._copy(value)]
        elif isinstance(current, set):
            current = current.copy()  # Create a copy to avoid modifying original
            current.add(self._copy(value))
            self._set_item_by_dot_path(key, current, cover=True)
            return
        elif isinstance(current, list):
            new_value = current.copy()  # Create a copy to avoid modifying original
            new_value.append(self._copy(value))
        else:
            new_value = [current, self._copy(value)]

        if isinstance(key, str) and "." in key:
            self._set_item_by_dot_path(key, new_value, cover=True)
        else:
            self._data[key] = new_value

    def extend(self, key: Any, values: Sequence[Any]):
        if isinstance(key, str) and "." in key:
            current = self._get_item_by_dot_path(key, inherit=False)
        else:
            current = self._data.get(key)

        if current is None:
            new_value = [self._copy(item) for item in values]
        elif isinstance(current, list):
            new_value = current.copy()  # Create a copy to avoid modifying original
            new_value.extend([self._copy(item) for item in values])
        else:
            new_value = [current]
            new_value.extend([self._copy(item) for item in values])

        if isinstance(key, str) and "." in key:
            self._set_item_by_dot_path(key, new_value, cover=True)
        else:
            self._data[key] = new_value

    def delete(self, key: Any):
        self.__delitem__(key)

    def namespace(self, namespace_path: str):
        return RuntimeDataNamespace(self, namespace_path)


class RuntimeDataNamespace:
    def __init__(self, root_runtime_data: RuntimeData, namespace: str):
        self.root = root_runtime_data
        self.namespace = namespace

    def __repr__(self) -> str:
        return f"RuntimeDataNamespace(root=RuntimeData(name={ str(self.root.name) }), data={ str(self.data) })"

    def __eq__(self, equal_target: Any) -> bool:
        return self.data == equal_target

    def __iter__(self) -> Iterator[Any]:
        """Make namespace iterable"""
        data = self.data
        if isinstance(data, dict):
            return iter(data)
        return iter({})

    def __len__(self) -> int:
        """Return the number of items in the namespace"""
        data = self.data
        if isinstance(data, dict):
            return len(data)
        return 0

    @property
    def data(self) -> Any:
        return self.root.get(self.namespace)

    def __getitem__(self, key: Any) -> Any:
        if isinstance(key, str) and "." in key:
            return self.root.get(f"{ self.namespace }.{ key }")
        else:
            ns_data = self.data
            if isinstance(ns_data, dict):
                return ns_data.get(key)
            return None

    def get(
        self,
        key: Any | None = None,
        default: T = None,
        inherit: bool = True,
    ) -> Any | T:
        if inherit:
            if key is not None:
                result = self[key]
            else:
                result = self.data
        else:
            if key is not None:
                ns_data = self.root.get(self.namespace, inherit=False)
                if isinstance(ns_data, dict):
                    if isinstance(key, str) and "." in key:
                        # Handle dot path within namespace
                        current = ns_data
                        for path_part in key.split("."):
                            if isinstance(current, dict) and path_part in current:
                                current = current[path_part]
                            else:
                                return default
                        return current
                    else:
                        return ns_data.get(key)
                return None
            else:
                result = self.root.get(self.namespace, inherit=False)
        return result if result is not None else default

    def keys(self):
        ns = self.data
        if isinstance(ns, dict):
            return ns.keys()
        return {}.keys()

    def values(self):
        ns = self.data
        if isinstance(ns, dict):
            return ns.values()
        return {}.values()

    def items(self):
        ns = self.data
        if isinstance(ns, dict):
            return ns.items()
        return {}.items()

    def __setitem__(self, key: Any, value: Any):
        if isinstance(key, str) and "." in key:
            return self.root.set(f"{self.namespace}.{key}", value)
        else:
            # Ensure namespace exists
            if self.root.get(self.namespace, inherit=False) is None:
                self.root._data[self.namespace] = {}

            self.root.set(f"{self.namespace}.{key}", value)

    def __delitem__(self, key: Any):
        if isinstance(key, str) and "." in key:
            del self.root[f"{ self.namespace }.{ key }"]
        else:
            ns = self.root.get(self.namespace, inherit=False)
            if isinstance(ns, dict) and key in ns:
                del ns[key]
                self.root._data[self.namespace] = ns

    def pop(self, key: str, default: Any = None) -> Any:
        if isinstance(key, str) and "." in key:
            return self.root.pop(f"{ self.namespace }.{ key }", default)
        else:
            ns = self.root.get(self.namespace, inherit=False)
            if isinstance(ns, dict) and key in ns:
                val = ns.pop(key)
                self.root._data[self.namespace] = ns
                return val
            return default

    def clear(self):
        self.root._data[self.namespace] = {}

    def __contains__(self, key: Any) -> bool:
        return key in self.keys()

    def set(self, key: Any, value: Any):
        return self.__setitem__(key, value)

    def setdefault(self, key: Any, value: Any, *, inherit: bool = True):
        if self.get(key, inherit=inherit) is None:
            self.set(key, value)
        return self.get(key, inherit=inherit)

    def update(self, new: dict[Any, Any]):
        for key, value in new.items():
            self.set(key, value)

    def append(self, key: str, value: Any):
        full_key = f"{ self.namespace }.{ key }"
        self.root.append(full_key, value)

    def extend(self, key: str, values: Sequence[Any]):
        full_key = f"{ self.namespace }.{ key }"
        self.root.extend(full_key, values)

    def delete(self, key: Any):
        self.__delitem__(key)
